#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Worlds are the basic environments which define how agents interact with one another.

    ``World(object)`` provides a generic parent class, including ``__enter__``
    and ``__exit__`` statements which allow you to guarantee that the shutdown
    method is called.

    ``DialogPartnerWorld(World)`` provides a two-agent turn-based dialog setting.

    ``MultiAgentDialogWorld(World)`` provides a multi-agent setting.

    ``MultiWorld(World)`` creates a set of environments (worlds) for the same agent
    to multitask over, a different environment will be chosen per episode.

    ``BatchWorld(World)`` is a container for doing minibatch training over a world by
    collecting batches of N copies of the environment (each with different state).


All worlds are initialized with the following parameters:

    ``opt`` -- contains any options needed to set up the agent. This generally contains
        all command-line arguments recognized from core.params, as well as other
        options that might be set through the framework to enable certain modes.
    ``agents`` -- the set of agents that should be attached to the world,
        e.g. for DialogPartnerWorld this could be the teacher (that defines the
        task/dataset) and the learner agent. This is ignored in the case of
        sharing, and the shared parameter is used instead to initalize agents.
    ``shared`` (optional) -- if not None, contains any shared data used to construct
        this particular instantiation of the world. This data might have been
        initialized by another world, so that different agents can share the same
        data (possibly in different Processes).
"""

import copy
import random

from typing import List, Dict, Union

from parlai.core.agents import create_agents_from_shared
from parlai.core.loader import load_task_module, load_world_module
from parlai.core.metrics import aggregate_named_reports
from parlai.core.opt import Opt
from parlai.core.teachers import Teacher, create_task_agent_from_taskname
from parlai.utils.data import DatatypeHelper
from parlai.utils.misc import Timer, display_messages
from parlai.tasks.tasks import ids_to_tasks
import parlai.utils.logging as logging


def validate(observation):
    """
    Make sure the observation table is valid, or raise an error.
    """
    if observation is not None and isinstance(observation, dict):
        return observation
    else:
        raise RuntimeError('Must return dictionary or Message object from act().')


class World(object):
    """
    Empty parent providing null definitions of API functions for Worlds.

    All children can override these to provide more detailed functionality.
    """

    def __init__(self, opt: Opt, agents=None, shared=None):
        self.id = opt['task']
        self.opt = copy.deepcopy(opt)
        if shared:
            # Create agents based on shared data.
            self.agents = create_agents_from_shared(shared['agents'])
        else:
            # Add passed in agents to world directly.
            self.agents = agents
        self.max_exs = None
        self.total_exs = 0
        self.total_epochs = 0
        self.total_parleys = 0
        self.time = Timer()

    def parley(self):
        """
        Perform one step of actions for the agents in the world.

        This is empty in the base class.
        """
        # TODO: mark as abstract?
        pass

    def getID(self):
        """
        Return the name of the world, typically the task the world encodes.
        """
        return self.id

    def display(self):
        """
        Return a string describing the current state of the world.

        Useful for monitoring and debugging. By default, display the messages between
        the agents.
        """
        if not hasattr(self, 'acts'):
            return ''
        return display_messages(
            self.acts,
            ignore_agent_reply=self.opt.get('ignore_agent_reply', False),
            add_fields=self.opt.get('display_add_fields', ''),
            prettify=self.opt.get('display_prettify', False),
            max_len=self.opt.get('max_display_len', 1000),
            verbose=self.opt.get('verbose', False),
        )

    def episode_done(self):
        """
        Whether the episode is done or not.
        """
        return False

    def epoch_done(self):
        """
        Whether the epoch is done or not.

        Not all worlds have the notion of an epoch, but this is useful for fixed
        training, validation or test sets.
        """
        return False

    def share(self):
        """
        Share the world.
        """
        shared_data = {}
        shared_data['world_class'] = type(self)
        shared_data['opt'] = self.opt
        shared_data['agents'] = self._share_agents()
        return shared_data

    def clone(self):
        """
        Create a duplicate of the world.
        """
        return type(self)(opt=self.opt, agents=None, shared=self.share())

    def _share_agents(self):
        """
        Create shared data for agents.

        Allows other classes to create the same agents without duplicating the data
        (i.e. sharing parameters).
        """
        if not hasattr(self, 'agents'):
            return None
        shared_agents = [a.share() for a in self.agents]
        return shared_agents

    def get_agents(self):
        """
        Return the list of agents.
        """
        return self.agents

    def get_task_agent(self):
        """
        Return task agent, if applicable.
        """
        raise NotImplementedError('Implement in subworld')

    def get_model_agent(self):
        """
        Return model agent, if applicable.
        """
        raise NotImplementedError('Implement in subworld')

    def get_acts(self):
        """
        Return the last act of each agent.
        """
        return self.acts

    def get_time(self):
        """
        Return total training time.
        """
        return self.time.time()

    def get_total_exs(self):
        """
        Return total amount of examples seen by world.
        """
        return self.total_exs

    def get_total_epochs(self):
        """
        Return total amount of epochs on which the world has trained.
        """
        return self.total_epochs

    def get_total_parleys(self):
        """
        Return total number of parleys.
        """
        return self.total_parleys

    def __enter__(self):
        """
        Empty enter provided for use with ``with`` statement.

        e.g:

        .. code-block:: python

            with World() as world:
                for n in range(10):
                    n.parley()
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        After ``with`` statement, call shutdown.
        """
        self.shutdown()
        return False

    def num_examples(self):
        """
        Return the number of examples.

        Always 0 in the abstract world.
        """
        # TODO: mark as abstract?
        return 0

    def num_episodes(self):
        """
        Return the number of episodes.

        Always 0 in the abstract world.
        """
        # TODO: mark as abstract?
        return 0

    def reset(self):
        """
        Reset all agents in the world, and world statistics.
        """
        for a in self.agents:
            a.reset()
        self.max_exs = None
        self.total_exs = 0
        self.total_epochs = 0
        self.total_parleys = 0
        self.time.reset()

    def reset_metrics(self):
        """
        Reset metrics for all agents.
        """
        for a in self.agents:
            a.reset_metrics()

    def shutdown(self):
        """
        Perform any cleanup, if appropriate.
        """
        pass

    def update_counters(self):
        """
        Update how many epochs have completed.
        """
        self.total_parleys += 1
        if self.max_exs is None:
            if 'num_epochs' in self.opt and self.opt['num_epochs'] > 0:
                if self.num_examples:
                    self.max_exs = self.num_examples() * self.opt['num_epochs']
                else:
                    self.max_exs = -1
            else:
                self.max_exs = -1
        # when we know the size of the data
        if self.max_exs > 0 or self.num_examples():
            self.total_epochs = (
                self.total_parleys * self.opt.get('batchsize', 1) / self.num_examples()
            )
        # when we do not know the size of the data
        else:
            if self.epoch_done():
                self.total_epochs += 1


class DialogPartnerWorld(World):
    """
    Simple world for two agents communicating synchronously.

    This basic world switches back and forth between two agents, giving each agent one
    chance to speak per turn and passing that back to the other one.
    """

    def __init__(self, opt: Opt, agents=None, shared=None):
        if not ((agents is not None) ^ (shared is not None)):
            raise ValueError('You must supply either agents or shared, but not both.')
        super().__init__(opt)
        if shared:
            # Create agents based on shared data.
            self.agents = create_agents_from_shared(shared['agents'])
        else:
            if len(agents) != 2:
                raise RuntimeError('There must be exactly two agents for this world.')
            # Add passed in agents directly.
            self.agents = agents
        self.acts = [None] * len(self.agents)
        if self.agents is not None and len(self.agents) > 0:
            # Name the world after the first agent.
            self.id = self.get_task_agent().getID()

    def get_task_agent(self):
        """
        Return task agent.
        """
        return self.get_agents()[0]

    def get_model_agent(self):
        """
        Return model agent, if applicable.
        """
        return self.get_agents()[1]

    def parley(self):
        """
        Agent 0 goes first.

        Alternate between the two agents.
        """
        acts = self.acts
        agents = self.agents
        acts[0] = agents[0].act()
        agents[1].observe(validate(acts[0]))
        acts[1] = agents[1].act()
        agents[0].observe(validate(acts[1]))
        self.update_counters()

    def episode_done(self):
        """
        Only the first agent indicates when the episode is done.
        """
        if self.acts[0] is not None:
            return self.acts[0].get('episode_done', False)
        else:
            return False

    def epoch_done(self):
        """
        Only the first agent indicates when the epoch is done.
        """
        return self.get_task_agent().epoch_done()

    def report(self):
        """
        Report all metrics of all subagents.
        """
        from parlai.core.metrics import Metric, LegacyMetric

        metrics = {}
        for a in self.agents:
            if hasattr(a, 'report'):
                m = a.report()
                for k, v in m.items():
                    if not isinstance(v, Metric):
                        v = LegacyMetric(v)
                    if k not in metrics:
                        # first agent gets priority in settings values for keys
                        # this way model can't e.g. override accuracy to 100%
                        metrics[k] = v
        if metrics and 'exs' in metrics:
            self.total_exs += metrics['exs'].value()
        return metrics

    def num_examples(self):
        """
        Return number of examples.
        """
        if hasattr(self.get_task_agent(), 'num_examples'):
            return self.get_task_agent().num_examples()
        return 0

    def num_episodes(self):
        """
        Return number of episodes.
        """
        if hasattr(self.get_task_agent(), 'num_episodes'):
            return self.get_task_agent().num_episodes()
        return 0

    def shutdown(self):
        """
        Shutdown each agent.
        """
        for a in self.agents:
            a.shutdown()

    def update_counters(self):
        """
        Ensure all counters are synchronized across threads.
        """
        super().update_counters()
        for a in self.agents:
            if hasattr(a, 'update_counters'):
                a.update_counters()


class MultiAgentDialogWorld(World):
    """
    Basic world where each agent gets a turn in a round-robin fashion.

    Each agent receives as input the actions of all other agents since its last `act()`.
    """

    def __init__(self, opt: Opt, agents, shared=None):
        super().__init__(opt)
        if shared:
            # Create agents based on shared data.
            self.agents = create_agents_from_shared(shared['agents'])
        else:
            # Add passed in agents directly.
            self.agents = agents
        self.acts = [None] * len(self.agents)

    def parley(self):
        """
        Perform a turn for every agent.

        For each agent, get an observation of the last action each of the other agents
        took. Then take an action yourself.
        """
        acts = self.acts
        for index, agent in enumerate(self.agents):
            acts[index] = agent.act()
            for other_agent in self.agents:
                if other_agent != agent:
                    other_agent.observe(validate(acts[index]))
        self.update_counters()

    def get_task_agent(self):
        """
        Return task agent.
        """
        return self.get_agents()[0]

    def get_model_agent(self):
        """
        Return model agent.
        """
        return self.get_agents()[1]

    def epoch_done(self):
        """
        Return if the epoch is done for any subagent.
        """
        done = False
        for a in self.agents:
            if a.epoch_done():
                done = True
        return done

    def episode_done(self):
        """
        Return if the episode is done for any subagent.
        """
        done = False
        for a in self.agents:
            if a.episode_done():
                done = True
        return done

    def report(self):
        """
        Report metrics for all subagents.
        """
        metrics = {}
        for a in self.agents:
            if hasattr(a, 'report'):
                m = a.report()
                for k, v in m.items():
                    if k not in metrics:
                        # first agent gets priority in settings values for keys
                        # this way model can't e.g. override accuracy to 100%
                        metrics[k] = v
        if metrics and 'exs' in metrics:
            self.total_exs += metrics['exs'].value()
        return metrics

    def shutdown(self):
        """
        Shutdown each agent.
        """
        for a in self.agents:
            a.shutdown()


class MultiWorld(World):
    """
    Container for multiple worlds.

    Container for a set of worlds where each world gets a turn in a round-robin fashion.
    The same user_agents are placed in each, though each world may contain additional
    agents according to the task that world represents.
    """

    def __init__(self, opt: Opt, agents=None, shared=None, default_world=None):
        super().__init__(opt)
        self.worlds: List[World] = []
        for index, k in enumerate(opt['task'].split(',')):
            k = k.strip()
            if k:
                if shared:
                    # Create worlds based on shared data.
                    s = shared['worlds'][index]
                    self.worlds.append(s['world_class'](s['opt'], None, s))
                else:
                    # Agents are already specified.
                    opt_singletask = copy.deepcopy(opt)
                    opt_singletask['task'] = k
                    self.worlds.append(
                        create_task_world(
                            opt_singletask, agents, default_world=default_world
                        )
                    )
        self.world_idx = -1
        self.new_world = True
        self.parleys = -1
        # Check to see if we are training
        self.is_training = DatatypeHelper.is_training(opt.get('datatype'))
        # Make multi-task task probabilities.
        self.cum_task_weights = [1] * len(self.worlds)
        self.task_choices = range(len(self.worlds))
        weights = self.opt.get('multitask_weights', [1])
        if weights == 'stochastic':
            weights = [w.num_episodes() for w in self.worlds]
        sum = 0
        for i in self.task_choices:
            if len(weights) > i:
                weight = weights[i]
            else:
                weight = 1
            self.cum_task_weights[i] = weight + sum
            sum += weight
        task_ids: Dict[str, Teacher] = {}
        # Having overlap in teacher ids will cause issues for metrics aggregation.
        for each_world in self.worlds:
            world_id = each_world.getID()
            if world_id in task_ids:
                raise AssertionError(
                    '{} and {} teachers have overlap in id {}.'.format(
                        task_ids[world_id],
                        each_world.get_agents()[0].__class__,
                        world_id,
                    )
                )
            else:
                task_ids[world_id] = each_world.get_task_agent()

    def num_examples(self):
        """
        Return sum of each subworld's number of examples.
        """
        if not hasattr(self, 'num_exs'):
            worlds_num_exs = [w.num_examples() for w in self.worlds]
            if any(num is None for num in worlds_num_exs):
                self.num_exs = None
            else:
                self.num_exs = sum(worlds_num_exs)
        return self.num_exs

    def num_episodes(self):
        """
        Return sum of each subworld's number of episodes.
        """
        if not hasattr(self, 'num_eps'):
            worlds_num_eps = [w.num_episodes() for w in self.worlds]
            if any(num is None for num in worlds_num_eps):
                self.num_eps = None
            else:
                self.num_eps = sum(worlds_num_eps)
        return self.num_eps

    def get_agents(self):
        """
        Return the agents in the *current* subworld.
        """
        return self.worlds[self.world_idx].get_agents()

    def get_task_agent(self):
        """
        Not possible/well-defined in this setting.
        """
        return self.worlds[self.world_idx].get_task_agent()

    def get_model_agent(self):
        """
        Not implemented.
        """
        return self.worlds[self.world_idx].get_model_agent()

    def get_acts(self):
        """
        Return the acts in the *current* subworld.
        """
        return self.worlds[self.world_idx].get_acts()

    def share(self):
        """
        Share all the subworlds.
        """
        shared_data = {}
        shared_data['world_class'] = type(self)
        shared_data['opt'] = self.opt
        shared_data['worlds'] = [w.share() for w in self.worlds]
        return shared_data

    def epoch_done(self):
        """
        Return if *all* the subworlds are done.
        """
        for t in self.worlds:
            if not t.epoch_done():
                return False
        return True

    def parley_init(self):
        """
        Update the current subworld.

        If we are in the middle of an episode, keep the same world and finish this
        episode. If we have finished this episode, pick a new world (either in a random
        or round-robin fashion).
        """
        self.parleys = self.parleys + 1
        if self.world_idx >= 0 and self.worlds[self.world_idx].episode_done():
            self.new_world = True
        if self.new_world:
            self.new_world = False
            self.parleys = 0
            if self.is_training:
                # select random world
                self.world_idx = random.choices(
                    self.task_choices, cum_weights=self.cum_task_weights
                )[0]
            else:
                # do at most one full loop looking for unfinished world
                for _ in range(len(self.worlds)):
                    self.world_idx = (self.world_idx + 1) % len(self.worlds)
                    if not self.worlds[self.world_idx].epoch_done():
                        # if this world has examples ready, break
                        break

    def parley(self):
        """
        Parley the *current* subworld.
        """
        self.parley_init()
        self.worlds[self.world_idx].parley()
        self.update_counters()

    def display(self):
        """
        Display all subworlds.
        """
        if self.world_idx != -1:
            s = ''
            w = self.worlds[self.world_idx]
            if self.parleys == 0:
                s = '[world ' + str(self.world_idx) + ':' + w.getID() + ']\n'
            s = s + w.display()
            return s
        else:
            return ''

    def report(self):
        """
        Report aggregate metrics across all subworlds.
        """
        metrics = aggregate_named_reports(
            {w.getID(): w.report() for w in self.worlds},
            micro_average=self.opt.get('aggregate_micro', False),
        )
        if 'exs' in metrics:
            self.total_exs += metrics['exs'].value()
        return metrics

    def reset(self):
        """
        Reset all subworlds.
        """
        for w in self.worlds:
            w.reset()

    def reset_metrics(self):
        """
        Reset metrics in all subworlds.
        """
        for w in self.worlds:
            w.reset_metrics()

    def update_counters(self):
        super().update_counters()
        for w in self.worlds:
            w.update_counters()


def _override_opts_in_shared(table, overrides):
    """
    Override all shared dicts.

    Looks recursively for ``opt`` dictionaries within shared dict and overrides any key-
    value pairs with pairs from the overrides dict.
    """
    if 'opt' in table:
        # change values if an 'opt' dict is available
        for k, v in overrides.items():
            table['opt'][k] = v
    for k, v in table.items():
        # look for sub-dictionaries which also might contain an 'opt' dict
        if type(v) == dict and k != 'opt' and 'opt' in v:
            _override_opts_in_shared(v, overrides)
        elif type(v) == list:
            for item in v:
                if type(item) == dict and 'opt' in item:
                    # if this is a list of agent shared dicts, we want to iterate
                    _override_opts_in_shared(item, overrides)
                else:
                    # if this is e.g. list of candidate strings, stop right away
                    break
    return table


class BatchWorld(World):
    """
    BatchWorld contains many copies of the same world.

    Create a separate world for each item in the batch, sharing
    the parameters for each.

    The underlying world(s) it is batching can be either
    ``DialogPartnerWorld``, ``MultiAgentWorld``, or ``MultiWorld``.
    """

    def __init__(self, opt: Opt, world):
        super().__init__(opt)
        self.opt = opt
        self.random = opt.get('datatype', None) == 'train'
        self.world = world
        self.worlds: List[World] = []
        for i in range(opt['batchsize']):
            # make sure that any opt dicts in shared have batchindex set to i
            # this lets all shared agents know which batchindex they have,
            # which is needed for ordered data (esp valid/test sets)
            shared = world.share()
            shared['batchindex'] = i
            for agent_shared in shared.get('agents', ''):
                agent_shared['batchindex'] = i
            # TODO: deprecate override_opts
            _override_opts_in_shared(shared, {'batchindex': i})
            self.worlds.append(shared['world_class'](opt, None, shared))
        self.batch_observations = [None] * len(self.world.get_agents())
        self.first_batch = None
        self.acts = [None] * len(self.world.get_agents())

    def batch_observe(self, index, batch_actions, index_acting):
        """
        Observe corresponding actions in all subworlds.
        """
        batch_observations = []
        for i, w in enumerate(self.worlds):
            agents = w.get_agents()
            observation = None
            if batch_actions[i] is None:
                # shouldn't send None, should send empty observations
                batch_actions[i] = [{}] * len(self.worlds)

            if hasattr(w, 'observe'):
                # The world has its own observe function, which the action
                # first goes through (agents receive messages via the world,
                # not from each other).
                observation = w.observe(agents[index], validate(batch_actions[i]))
            else:
                observation = validate(batch_actions[i])

            if index == index_acting:
                # self_observe is distinguished from a normal observe
                if hasattr(agents[index], 'self_observe'):
                    agents[index].self_observe(observation)
            else:
                observation = agents[index].observe(observation)

            # TODO: not so sure about this...
            if observation is None:
                raise ValueError('Agents should return what they observed.')
            batch_observations.append(observation)
        return batch_observations

    def batch_act(self, agent_idx, batch_observation):
        """
        Act in all subworlds.
        """
        # Given batch observation, do update for agents[index].
        # Call update on agent
        a = self.world.get_agents()[agent_idx]
        if hasattr(a, 'batch_act'):
            batch_actions = a.batch_act(batch_observation)
            # Store the actions locally in each world.
            for i, w in enumerate(self.worlds):
                acts = w.get_acts()
                acts[agent_idx] = batch_actions[i]
        else:
            # Reverts to running on each individually.
            batch_actions = []
            for w in self.worlds:
                agents = w.get_agents()
                acts = w.get_acts()
                acts[agent_idx] = agents[agent_idx].act()
                batch_actions.append(acts[agent_idx])
        return batch_actions

    def parley(self):
        """
        Parley in all subworlds.

        Usually with ref:`batch_act` and ref:`batch_observe`.
        """
        # Collect batch together for each agent, and do update.
        # Assumes DialogPartnerWorld, MultiAgentWorld, or MultiWorlds of them.
        num_agents = len(self.world.get_agents())
        batch_observations = self.batch_observations

        if hasattr(self.world, 'parley_init'):
            for w in self.worlds:
                w.parley_init()

        for agent_idx in range(num_agents):
            # The agent acts.
            batch_act = self.batch_act(agent_idx, batch_observations[agent_idx])
            self.acts[agent_idx] = batch_act
            # We possibly execute this action in the world.
            if hasattr(self.world, 'execute'):
                for w in self.worlds:
                    w.execute(w.agents[agent_idx], batch_act[agent_idx])
            # All agents (might) observe the results.
            for other_index in range(num_agents):
                obs = self.batch_observe(other_index, batch_act, agent_idx)
                if obs is not None:
                    batch_observations[other_index] = obs
        self.update_counters()

    def display(self):
        """
        Display the full batch.
        """
        s = "[--batchsize " + str(len(self.worlds)) + "--]\n"
        for i, w in enumerate(self.worlds):
            s += "[batch world " + str(i) + ":]\n"
            s += w.display() + '\n'
        s += "[--end of batch--]"
        return s

    def num_examples(self):
        """
        Return the number of examples for the root world.
        """
        return self.world.num_examples()

    def num_episodes(self):
        """
        Return the number of episodes for the root world.
        """
        return self.world.num_episodes()

    def get_total_exs(self):
        """
        Return the total number of processed episodes in the root world.
        """
        return self.world.get_total_exs()

    def getID(self):
        """
        Return the ID of the root world.
        """
        return self.world.getID()

    def get_agents(self):
        """
        Return the agents of the root world.
        """
        return self.world.get_agents()

    def get_task_agent(self):
        """
        Return task agent of the root world.
        """
        return self.world.get_task_agent()

    def get_model_agent(self):
        """
        Return model agent of the root world.
        """
        return self.world.get_model_agent()

    def episode_done(self):
        """
        Return whether the episode is done.

        A batch world is never finished, so this always returns `False`.
        """
        return False

    def epoch_done(self):
        """
        Return if the epoch is done in the root world.
        """
        # first check parent world: if it says it's done, we're done
        if self.world.epoch_done():
            return True
        # otherwise check if all shared worlds are done
        for world in self.worlds:
            if not world.epoch_done():
                return False
        return True

    def report(self):
        """
        Report metrics for the root world.
        """
        return self.world.report()

    def reset(self):
        """
        Reset the root world, and all copies.
        """
        self.world.reset()
        for w in self.worlds:
            w.reset()

    def reset_metrics(self):
        """
        Reset metrics in the root world.
        """
        self.world.reset_metrics()

    def shutdown(self):
        """
        Shutdown each world.
        """
        for w in self.worlds:
            w.shutdown()
        self.world.shutdown()


class DynamicBatchWorld(World):
    def __init__(self, opt: Opt, world: Union[DialogPartnerWorld, MultiWorld]):
        super().__init__(opt)
        self.opt = opt

        # agents is a placeholder just for super.reset
        self.agents = []

        # check some assumptions
        if isinstance(world, (BatchWorld, MultiAgentDialogWorld)):
            raise TypeError(
                'World must be a DialogPartnerWorld or a '
                'MultiWorld of DialogPartnerWorld'
            )

        if len(world.get_agents()) != 2:
            raise AssertionError(
                "Dynamic batch only works in a fixed dialog world with two agents."
            )

        if not hasattr(world.get_model_agent(), 'batch_act'):
            raise TypeError("Model agent doesn't have batch_act.")

        self.truncate = opt.get('text_truncate', None) or opt.get('truncate', None)
        self.l_truncate = opt.get('label_truncate', None) or opt.get('truncate', None)
        if self.truncate is None or self.truncate < 0:
            raise ValueError(
                'You must use --text-truncate or --truncate in order to use '
                'dynamic batching.'
            )

        # size of the buffer we will use to find worlds
        self._BUFFER_SIZE = 1021  # chosen as a prime number

        if opt['dynamic_batching'] == 'full':
            # full dynamic batching, we can grow our batchsize
            self.max_batch_size = self._BUFFER_SIZE
        else:
            # simple batchsort
            self.max_batch_size = opt['batchsize']

        # TODO: check to ensure the agent has self_observe
        self.world = world
        # TODO: maybe generalize this
        self.max_words = (self.l_truncate + self.truncate) * opt['batchsize']

        # buffer worlds
        self.worlds = [world.clone() for _ in range(self._BUFFER_SIZE)]

        self.reset()

    def reset(self):
        super().reset()
        self._task_acts = [None for _ in range(self._BUFFER_SIZE)]
        self._obs = [None for _ in range(self._BUFFER_SIZE)]
        self._scores = [None for _ in range(self._BUFFER_SIZE)]
        self.acts = [None, None]

        self.number_parleys = 0
        self.total_exs = 0
        self.world.reset()
        self.rng = random.Random(4)
        for w in self.worlds:
            w.reset()

    def reset_metrics(self):
        super().reset_metrics()
        self.world.reset_metrics()
        for w in self.worlds:
            w.reset_metrics()

    def epoch_done(self):
        return (
            self.world.epoch_done()
            or all(w.epoch_done() for w in self.worlds)
            and all(s is None for s in self._scores)
        )

    def num_examples(self):
        return self.world.num_examples()

    def num_episodes(self):
        return self.world.num_episodes()

    def _ceil(self, n):
        """
        Round to the nearest multiple of 8.

        TensorCores only work when a tensor is a multiple of 8 in almost all
        dimensions. This means all examples cost is related to their nearest
        multiple of 8.

        See https://devblogs.nvidia.com/programming-tensor-cores-cuda-9/ for
        more information.
        """
        # round up to r, all things are equal
        from parlai.utils.torch import FP16_PAD_SIZE

        return ((n + FP16_PAD_SIZE - 1) // FP16_PAD_SIZE) * FP16_PAD_SIZE

    def _score(self, obs):
        if 'text_vec' in obs:
            # note that all examples have a cost that is based on their
            # nearest multiple of 4. We can therefore mix-and-match
            # anything with the same cost for increased stochasticity,
            # while not really wasting much padding.
            return tuple(
                self._ceil(len(obs[key]))
                for key in ['text_vec', 'labels_vec', 'eval_labels_vec']
                if key in obs
            )
        else:
            return None

    def parley(self):
        # first make sure that all the worlds are processed in the queue
        indices = []
        for i in range(self._BUFFER_SIZE):
            if self._scores[i] is not None:
                indices.append(i)
                continue
            if self.worlds[i].epoch_done():
                continue

            if hasattr(self.world, 'parley_init'):
                self.worlds[i].parley_init()

            act = self.worlds[i].get_task_agent().act()

            # we log the task act and the index of the act
            # in the buffer for world logging purposes
            self._task_acts[i] = act  # for world logging
            self._task_acts[i]['dyn_batch_idx'] = i

            obs = self.worlds[i].get_model_agent().observe(act)
            self._obs[i] = obs

            self._scores[i] = self._score(obs)
            if self._scores[i] is not None:
                indices.append(i)

        # quick invariant checks
        assert len(indices) != 0, "DynamicBatchWorld ran out of data!"
        assert not any(self._scores[i] is None for i in indices)

        # sort all the indices by their score, so that we can find similarly lengthed
        # items in O(1)
        indices = sorted(indices, key=lambda i: self._scores[i] + (self.rng.random(),))

        # now let's build the batch
        batch = []

        # start with a random item. indices_idx is the lookup into indices, but
        # index is the actual world.
        width = 0
        indices_idx = random.randint(0, len(indices) - 1)

        # we picked a random spot, but we can get better packing if we start at
        # the last example with the same score, since we always move down to
        # smaller examples.
        while indices_idx < len(indices) - 1 and (
            sum(self._scores[indices[indices_idx]])
            == sum(self._scores[indices[indices_idx + 1]])
        ):
            indices_idx += 1

        # quit early if we eat our full buffer
        while indices:
            index = indices[indices_idx]
            this_width = self._ceil(sum(self._scores[index]))
            new_width = max(width, this_width)
            # compute the cost of the new batch
            new_bsz = len(batch) + 1
            new_words = new_width * new_bsz
            if new_words <= self.max_words and new_bsz <= self.max_batch_size:
                # cool, this one fits, let's add it
                width = new_width
                batch.append(index)
                indices.pop(indices_idx)
                indices_idx = max(indices_idx - 1, 0)
            else:
                # we'd overfill our buffer, give up
                break

        # Always have a batch size that's a multiple of 4, for fp16's sake.
        while len(batch) > 4 and len(batch) % 4 != 0:
            # pop off the shortest one. it's easiest to pack in later
            batch.pop(-1)

        # double check our assumed invariant
        assert self._ceil(width) * len(batch) <= self.max_words
        assert len(batch) > 0
        assert len(batch) <= self.max_batch_size

        # great, this batch is good to go! let's run it!
        acts = self.world.get_model_agent().batch_act([self._obs[i] for i in batch])
        self.acts = [[self._task_acts[i] for i in batch], acts]
        # broadcast the results back to all the models
        for i, act in zip(batch, acts):
            # we need to make sure that the teachers saw the result
            self.worlds[i].get_task_agent().observe(act)
            # and that the agent copies saw their own voice
            self.worlds[i].get_model_agent().self_observe(act)
            # move these worlds forward
            act = self.worlds[i].get_task_agent().act()
            # we log the task act and the index of the act
            # in the buffer for world logging purposes
            self._task_acts[i] = act
            self._task_acts[i]['dyn_batch_idx'] = i
            # save the observations to form a batch
            obs = self.worlds[i].get_model_agent().observe(act)
            self._scores[i] = self._score(obs)
            self._obs[i] = obs

        # update metrics
        self.total_parleys += 1
        self.total_exs += len(batch)

    def get_total_exs(self):
        return self.total_exs

    def get_total_epochs(self):
        return self.total_exs / self.num_examples()

    def report(self):
        return self.world.report()


################################################################################
# Functions for creating tasks/worlds given options.
################################################################################
def _create_task_agents(opt: Opt):
    """
    Create task agent(s) for the given task name.

    It does this by calling the create_agent function in agents.py of the given task. If
    create_agents function does not exist, it just looks for the teacher (agent) class
    defined by the task name directly.  (This saves the task creator bothering to define
    the create_agents function when it is not needed.)
    """
    if opt.get('interactive_task', False) or opt.get('selfchat_task', False):
        # do not need task agents in interactive or self chat settings
        return []

    try:
        # Tries to call the create_agent function in agents.py
        my_module = load_task_module(opt['task'])
        task_agents = my_module.create_agents(opt)  # type: ignore
    except (ModuleNotFoundError, AttributeError):
        # Create_agent not found, so try to create the teacher directly.
        return create_task_agent_from_taskname(opt)
    if type(task_agents) != list:
        task_agents = [task_agents]
    return task_agents


def create_task_world(opt: Opt, user_agents, default_world=None):
    """
    Instantiate a world with the supplied options and user agents.

    (A world factory.)
    """
    task_agents = _create_task_agents(opt)
    world_class = load_world_module(
        opt['task'],
        interactive_task=opt.get('interactive_task', False),
        selfchat_task=opt.get('selfchat_task', False),
        num_agents=len(user_agents + task_agents),
        default_world=default_world,
    )

    return world_class(opt, task_agents + user_agents)


def create_task(opt: Opt, user_agents, default_world=None):
    """
    Create a world + task_agents (aka a task).

    Assuming ``opt['task']="task_dir:teacher_class:options"`` e.g. ``"babi:Task1k:1"``
    or ``"#babi-1k"`` or ``"#QA"``, see ``parlai/tasks/tasks.py`` and see
    ``parlai/tasks/task_list.py`` for list of tasks.
    """
    task = opt.get('task')
    if not task:
        raise RuntimeError(
            'No task specified. Please select a task with ' + '--task {task_name}.'
        )
    if type(user_agents) != list:
        user_agents = [user_agents]

    # Convert any hashtag task labels to task directory path names.
    # (e.g. "#QA" to the list of tasks that are QA tasks).
    opt = copy.deepcopy(opt)
    opt['task'] = ids_to_tasks(opt['task'])
    logging.info(f"creating task(s): {opt['task']}")

    if ',' not in opt['task']:
        # Single task
        world = create_task_world(opt, user_agents, default_world=default_world)
    else:
        # Multitask teacher/agent
        # TODO: remove and replace with multiteachers only?
        world = MultiWorld(opt, user_agents, default_world=default_world)

    if opt.get('batchsize', 1) > 1 and opt.get('dynamic_batching'):
        world = DynamicBatchWorld(opt, world)
    elif opt.get('batchsize', 1) > 1:
        # otherwise check if should use batchworld
        world = BatchWorld(opt, world)

    return world



#!/usr/bin/env python3

# Copyright (c) 2017-present, Moscow Institute of Physics and Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import os

RESOURCES = [
    DownloadableFile(
        'https://raw.githubusercontent.com/deepmipt/turing-data/master/data_1501534800.tar.gz',
        'data_1501534800.tar.gz',
        'f1a4e7ee7264220cef6bf067b77d6f501023877643e77516c7acd66fbcdf0aaf',
    )
]


def build(opt):
    data_path = os.path.join(opt['datapath'], 'DialogueQE')
    version = '1501534800'

    if not build_data.built(data_path, version_string=version):
        print('[building data: ' + data_path + ']')

        if build_data.built(data_path):
            build_data.remove_dir(data_path)
        build_data.make_dir(data_path)

        for downloadable_file in RESOURCES:
            downloadable_file.download_file(data_path)

        os.rename(
            os.path.join(data_path, 'data_train_' + version + '.json'),
            os.path.join(data_path, 'train.json'),
        )
        os.rename(
            os.path.join(data_path, 'data_test_' + version + '.json'),
            os.path.join(data_path, 'test.json'),
        )

        build_data.mark_done(data_path, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import math
import os
import random
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, TYPE_CHECKING

import numpy as np
from mephisto.operations.registry import register_mephisto_abstraction
from mephisto.abstractions.blueprint import SharedTaskState
from mephisto.abstractions.blueprints.static_react_task.static_react_blueprint import (
    StaticReactBlueprint,
    StaticReactBlueprintArgs,
)
from omegaconf import DictConfig

if TYPE_CHECKING:
    from mephisto.data_model.task import TaskRun


def get_task_path():
    return os.path.dirname(__file__)


STATIC_BLUEPRINT_TYPE = 'turn_annotations_static_blueprint'
STATIC_IN_FLIGHT_QA_BLUEPRINT_TYPE = 'turn_annotations_static_inflight_qa_blueprint'


@dataclass
class TurnAnnotationsStaticBlueprintArgs(StaticReactBlueprintArgs):
    _blueprint_type: str = STATIC_BLUEPRINT_TYPE
    _group: str = field(
        default="TurnAnnotationsStaticBlueprint",
        metadata={
            'help': """This task renders conversations from a file and asks for turn by turn annotations of them."""
        },
    )
    random_seed: int = field(
        default=42, metadata={"help": 'Seed for random operations'}
    )
    annotation_question: str = field(
        default='Does this comment require any annotations? (Check all that apply)',
        metadata={
            "help": "The string displayed above the checkboxes for each annotation in the task."
        },
    )
    subtasks_per_unit: int = field(
        default=-1, metadata={"help": "Number of subtasks/comparisons to do per unit"}
    )
    annotation_indices_jsonl: Optional[str] = field(
        default=None,
        metadata={
            "help": "Specify which utterance indices to annotate per conversation in a JSONL file. Must be same length as conversations data-jsonl file. See example file in task_config/annotation_indices_example.jsonl"
        },
    )
    ask_reason: bool = field(
        default=False,
        metadata={
            "help": "If we want to ask the crowdworker for a reason for each of their annotations in a text field"
        },
    )
    conversation_count: Optional[int] = field(
        default=None,
        metadata={
            "help": "Specify a positive integer if you want to use only the first N conversations in the data file"
        },
    )
    onboarding_data: str = field(
        default=os.path.join(get_task_path(), 'task_config/onboarding.json'),
        metadata={
            "help": "Path to data and answers for onboarding task in JSON format"
        },
    )
    annotation_buckets: Optional[str] = field(
        default=None,
        metadata={
            "help": "As per Turn Annotations task, path to annotation buckets which will be checkboxes in the frontend for worker to annotate an utterance. If none provided, no checkboxes."
        },
    )
    response_field: bool = field(
        default=False,
        metadata={
            "help": "If we want a freeform textbox input for the crowdworker to respond to the message."
        },
    )


@register_mephisto_abstraction()
class TurnAnnotationsStaticBlueprint(StaticReactBlueprint):
    """
    This Blueprint has a subtasks number option to combine multiple conversations into
    "sub-HITs".

    It also has options for the onboarding data answers and the annotation bucket
    definitions.
    """

    ArgsClass = TurnAnnotationsStaticBlueprintArgs
    BLUEPRINT_TYPE = STATIC_BLUEPRINT_TYPE

    def __init__(
        self, task_run: "TaskRun", args: "DictConfig", shared_state: "SharedTaskState"
    ):
        super().__init__(task_run, args=args, shared_state=shared_state)
        random.seed(self.args.blueprint.random_seed)
        np.random.seed(self.args.blueprint.random_seed)
        self.subtasks_per_unit = self.args.blueprint.subtasks_per_unit
        self.conversation_count = self.args.blueprint.conversation_count

        if self.subtasks_per_unit <= 0:
            raise Exception(
                f'subtasks_per_unit must be greater than zero but was {self.subtasks_per_unit}'
            )

        self.raw_data = self._initialization_data_dicts

        # Load from file if needed specifying which utterances within each
        # conversation to annotate
        self.annotation_indices = None
        if self.args.blueprint.annotation_indices_jsonl:
            self.annotation_indices = []
            with open(
                self.args.blueprint.annotation_indices_jsonl, "r", encoding="utf-8-sig"
            ) as f:
                line = f.readline()
                while line:
                    conversation_indices = json.loads(line)
                    self.annotation_indices.append(conversation_indices)
                    line = f.readline()
            if len(self.annotation_indices) != len(self.raw_data):
                raise Exception(
                    f'Cannot specify a different length of annotation indices ({len(self.annotation_indices)}) than conversations ({len(self.raw_data)}).'
                )
            # TODO: should check that utterances specified are all bot
            # utterances (agent_idx == 1)

        if self.conversation_count:
            self.raw_data = self.raw_data[: self.conversation_count]
            if self.annotation_indices:
                self.annotation_indices = self.annotation_indices[
                    : self.conversation_count
                ]

        # Reorganize the self-chat data
        self._initialization_data_dicts = self.process_data(
            self.raw_data, annotation_indices=self.annotation_indices
        )

        # Now chunk the data into groups of <num_subtasks>
        grouped_data = []
        logging.info(
            f'Raw data length: {len(self.raw_data)}. self.subtasks_per_unit: {self.subtasks_per_unit}'
        )
        for i in range(0, len(self._initialization_data_dicts), self.subtasks_per_unit):
            chunk = self._initialization_data_dicts[i : i + self.subtasks_per_unit]
            grouped_data.append(chunk)
        self._initialization_data_dicts = grouped_data
        # Last group may have less unless an exact multiple
        logging.info(
            f'Grouped data into {len(self._initialization_data_dicts)} tasks with {self.subtasks_per_unit} subtasks each.'
        )

    def get_frontend_args(self) -> Dict[str, Any]:
        """
        Specifies what options within a task_config should be forwarded to the client
        for use by the task's frontend.
        """

        with open(self.args.blueprint.onboarding_data, "r", encoding="utf-8-sig") as f:
            onboarding_data = json.loads(f.read())

        annotation_buckets = None
        if self.args.blueprint.annotation_buckets:
            with open(
                self.args.blueprint.annotation_buckets, "r", encoding="utf-8-sig"
            ) as f:
                annotation_buckets = json.loads(f.read())

        return {
            "task_description": self.args.task.get('task_description', None),
            "task_title": self.args.task.get('task_title', None),
            "annotation_question": self.args.blueprint.annotation_question,
            "onboarding_data": onboarding_data,
            "annotation_buckets": annotation_buckets,
            "ask_reason": self.args.blueprint.ask_reason,
            "response_field": self.args.blueprint.response_field,
            "frame_height": '100%',
            "num_subtasks": self.args.blueprint.subtasks_per_unit,
            "block_mobile": True,
        }

    def process_data(self, data_dicts, annotation_indices=None):
        """
        Override this in a subclass if you want to change how data is processed from
        input file before being sent to the frontend.
        """
        output = []
        total_annotation_count = 0
        for conv_idx, d in enumerate(data_dicts):
            if annotation_indices:
                total_annotation_count += len(annotation_indices[conv_idx])
                # We only want to show the conversation up to the last
                # utterance we need annotations on, b/c otherwise may confuse
                # or bias the turkers
                if len(annotation_indices[conv_idx]) > 1:
                    logging.info(
                        f'Splitting {len(annotation_indices[conv_idx])} separate problematic utterance annotations in the same conversation into two separate conversations for this task. This avoids biasing the turkers with utterances that may come after one of the annotations.'
                    )
                for a in annotation_indices[conv_idx]:
                    processed_dialog = self._process_conversation(d, [a])
                    output.append(processed_dialog)
            else:
                processed_dialog = self._process_conversation(
                    d, annotation_indices=None
                )
                output.append(processed_dialog)
        print(
            f'Processed {len(data_dicts)} total conversations into {len(output)} conversations to be used in crowdsourcing task with {total_annotation_count} total annotations.'
        )
        np.random.shuffle(output)
        return output

    def _process_conversation(self, d, annotation_indices: Optional[List[int]] = None):
        """
        Helper function for processing conversations.

        :param annotation_indices:
            Array of turn indices to annotate of the
            actual conversation not including the context [So 0 is the "Hi!" if
            that's the first non-context utterance of the conversation.] If this is not
            specified, annotate all bot turns.
        :return: modified dialogue object
        """
        new_dialogue = []
        if annotation_indices is not None:
            max_turn_to_show = max(annotation_indices)
        else:
            max_turn_to_show = None
        adjusted_turn_idx = 0
        for full_turn in d['dialog']:
            if len(full_turn) != 2:
                print(
                    f'Warning! Skipping incomplete conversation! full_turn was: {full_turn}'
                )
                continue
            if max_turn_to_show is not None and adjusted_turn_idx > max_turn_to_show:
                logging.info(
                    f'Skipping {adjusted_turn_idx}th utterance, b/c max_turn_to_show was {max_turn_to_show}.'
                )
                continue
            # If there is a persona, which is context as in the ConvAI2
            # task, we don't want to display the persona utterances
            if 'persona' not in full_turn[0]['text']:
                do_annotate = False
                new_dialogue.append(
                    {
                        'text': full_turn[0]['text'],
                        'agent_idx': 0,
                        'do_annotate': do_annotate,
                        'other_metadata': full_turn[0].get('other_metadata'),
                    }
                )
                adjusted_turn_idx += 1
            if 'persona' not in full_turn[1]['text']:
                if annotation_indices:
                    do_annotate = adjusted_turn_idx in annotation_indices
                else:
                    do_annotate = True
                    # Default to annotating all bot utterances
                new_dialogue.append(
                    {
                        'text': full_turn[1]['text'],
                        'agent_idx': 1,
                        'do_annotate': do_annotate,
                        'other_metadata': full_turn[1].get('other_metadata'),
                    }
                )
                adjusted_turn_idx += 1
        if max_turn_to_show is not None and adjusted_turn_idx < max_turn_to_show:
            raise Exception(
                f'Conversation had {adjusted_turn_idx} but max_turn_to_show was {max_turn_to_show}'
            )
        return new_dialogue


@dataclass
class TurnAnnotationsStaticInFlightQABlueprintArgs(TurnAnnotationsStaticBlueprintArgs):
    _blueprint_type: str = STATIC_IN_FLIGHT_QA_BLUEPRINT_TYPE
    _group: str = field(
        default="TurnAnnotationsStaticInFlightQABlueprint",
        metadata={
            'help': """This task mixes in a live onboarding as the last subtask (in addition to an onboarding at the start), and actually increases the number of subtasks per unit by 1."""
        },
    )
    onboarding_in_flight_data: str = field(
        default=os.path.join(get_task_path(), 'task_config/onboarding_in_flight.jsonl'),
        metadata={
            "help": "Path to data and answers for onboarding task in JSON-L format (one JSON object per line per onboarding)"
        },
    )


@register_mephisto_abstraction()
class TurnAnnotationsStaticInFlightQABlueprint(TurnAnnotationsStaticBlueprint):
    """
    This Blueprint mixes in a live onboarding as the last subtask (in addition to an
    onboarding at the start), and actually increases the number of subtasks per unit by
    1.
    """

    ArgsClass = TurnAnnotationsStaticInFlightQABlueprintArgs
    BLUEPRINT_TYPE = STATIC_IN_FLIGHT_QA_BLUEPRINT_TYPE

    def __init__(
        self, task_run: "TaskRun", args: "DictConfig", shared_state: "SharedTaskState"
    ):
        super().__init__(task_run, args=args, shared_state=shared_state)

        raw_qc_convos = []
        with open(self.args.blueprint.onboarding_in_flight_data, "r") as f:
            line = f.readline()
            while line:
                qc_convo = json.loads(line)
                raw_qc_convos.append(qc_convo)
                line = f.readline()
        # Annotate all the utterances in the quality controls
        # So annotation_indices=None here
        self.quality_control_convos = self.process_data(
            raw_qc_convos, annotation_indices=None
        )

        # Re-chunk the data to add a quality control convo as the last subtask
        # No shuffling of the data for reproducibility's sake
        # (quality control will always be last subtask)
        # TODO: I don't think we need to re-chunk this actually; just iterate
        # over the data and add the quality control task
        all_data = []
        for grp in self._initialization_data_dicts:
            all_data.extend(grp)

        grouped_data = []
        number_of_tasks = math.floor(len(all_data) / self.subtasks_per_unit)
        for i in range(0, number_of_tasks):
            data_index = i * self.subtasks_per_unit
            chunk = all_data[data_index : data_index + self.subtasks_per_unit]
            qc_convo_idx = i % len(self.quality_control_convos)
            chunk.append(self.quality_control_convos[qc_convo_idx])
            grouped_data.append(chunk)

        self._initialization_data_dicts = grouped_data
        self.subtasks_per_unit = len(chunk)

        print(
            f'{self.__class__.__name__}: Grouped data into {len(self._initialization_data_dicts)} tasks with {self.subtasks_per_unit} subtasks each (added in-flight qualification task).'
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import os
from parlai.tasks.personality_captions.download_images import download_images

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/image_chat/image_chat.tgz',
        'image_chat.tgz',
        'ad733e181de33f1085166bb7af17fcf228504bd48228ed8cc20c5e7a9fa5d259',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'image_chat')
    image_path = os.path.join(opt['datapath'], 'yfcc_images')
    version = '1.0'
    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        build_data.mark_done(dpath, version)

    if (
        not build_data.built(image_path, version)
        and not opt.get('yfcc_path')
        and opt.get('load_images')
    ):
        download_images(opt, task='image_chat')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# This file contains teacher agents for Natural Question dataset.
# General infomration about the dataset: https://ai.google.com/research/NaturalQuestions
# Details about its content: https://github.com/google-research-datasets/natural-questions

import copy
import os
import jsonlines
from tqdm import tqdm
from typing import List, Optional, Tuple

import parlai.utils.logging as logging
from parlai.core.teachers import ChunkTeacher
from .build import build, DATASET_NAME_LOCAL
from .utils.text_utils import simplify_nq_example

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser


def _count_lines_in_file(fname):
    num_lines = 0
    with open(fname, 'r') as fi:
        for _ in fi:
            num_lines += 1
    return num_lines


def _context_type_key(is_html):
    return 'document_html' if is_html else 'document_text'


def _create_long_answer_from_span_html(example):
    """
    Creates a list of long answer candidates, from their spans on the document.

    This function gets the full article from the input example dictionary (using
    key 'document_html'), then iterates through the long answer spans (from
    'long_answer_candidates' key) and creates a list of slices from the article,
    using the 'start_byte' and 'end_byte' values in the list of long answer
    candidate spans.

    Returns a list of long answers. Each long answer is a substring from the
        original HTML text.

    :param example: a dict that contains one example/entry from NQ dataset.
    """
    context_text = example[_context_type_key(is_html=True)].encode()
    candidate_long_answers = []
    for long_answer_span in example['long_answer_candidates']:
        start_index = long_answer_span['start_byte']
        end_index = long_answer_span['end_byte']
        answer = context_text[start_index:end_index].decode()
        candidate_long_answers.append(answer)
    return candidate_long_answers


def _create_long_answer_from_span_text(simplified_example):
    """
    Creates a list of long answer candidates, from their spans on the document.

    This function gets the full article from the input simplified example
    dictionary (using key 'document_text'), then iterates through the long
    answer spans (from 'long_answer_candidates' key) and creates a list of
    slices from the article, using the 'start_token' and 'end_token' values in
    the list of long answer candidate spans.

    Returns a list of long answers. Each long answer is a substring from the
        simplified HTML text.

    :param simplified_example: a dict that contains one simplified example/entry
        from NQ dataset.
    """
    context_text = simplified_example[_context_type_key(is_html=False)]
    candidate_long_answers = []
    splitted_tokens = context_text.split(' ')
    for long_answer_span in simplified_example['long_answer_candidates']:
        start_index = long_answer_span['start_token']
        end_index = long_answer_span['end_token']
        answer = ' '.join(splitted_tokens[start_index:end_index])
        candidate_long_answers.append(answer)
    return candidate_long_answers


class NaturalQuestionsTeacher(ChunkTeacher):
    """
    The base teacher class for Natural Questions dataset challenge.

    This class implements the core functionalities for other teachers. The other four
    variations of teachers are made by setting two object attributes (use_html,
    use_long_answer) to either True or False.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        nq = parser.add_argument_group('Natural Questions Teacher')
        nq.add_argument(
            '--use-html',
            type='bool',
            default=False,
            help='Use HTML for the context (does nothing if `use-context` is False)',
        )
        nq.add_argument(
            '--use-long-answer', type='bool', default=False, help='Use long answers'
        )
        nq.add_argument(
            '--use-context',
            type='bool',
            default=True,
            help='Include context blurb or not',
        )

    def __init__(self, opt, shared=None):
        build(opt)
        self.use_html = opt.get('use_html', False)
        self.use_long_answer = opt.get('use_long_answer', False)
        self.use_context = opt.get('use_context', False)
        self.id = 'natural_questions'
        self.opt = copy.deepcopy(opt)
        self.dtype = self.opt['datatype'].split(':')[0]
        self.dpath = os.path.join(self.opt['datapath'], DATASET_NAME_LOCAL, self.dtype)
        self.n_samples = None
        super().__init__(self.opt, shared)

    def _simplify(self, example):
        if self.use_html:
            return example
        return simplify_nq_example(example)

    def _get_data_folder(self):
        return self.dpath

    def get_fold_chunks(self, opt) -> List[int]:
        if 'train' == self.dtype:
            return list(range(50))
        elif 'valid' == self.dtype:
            return list(range(5))
        raise ValueError(f'Invalid data type: "{self.dtype}"')

    def get_num_samples(self, opt) -> Tuple[int, int]:
        if self.n_samples:
            return self.n_samples
        logging.log(f'Counting the number of samples in {self.dtype}')
        files = os.listdir(self.dpath)
        n_samples = 0
        for fname in tqdm(files):
            if fname.startswith('.'):  # some of the OS specific files
                continue
            n_samples += _count_lines_in_file(os.path.join(self.dpath, fname))
        logging.info(f'{n_samples} examples found in {self.dtype} dataset.')
        self.n_samples = (n_samples, n_samples)
        return self.n_samples

    def _get_candidate_long_answers(self, example):
        if self.use_html:
            return _create_long_answer_from_span_html(example)
        else:
            return _create_long_answer_from_span_text(example)

    def _get_short_answers(self, example):
        context = example[_context_type_key(self.use_html)]
        if self.use_html:
            offset_unit = 'byte'
            context = context.encode()
        else:
            offset_unit = 'token'
            context = context.split(' ')

        short_answers = []
        for annotation in example['annotations']:
            if 'short_answers' in annotation and annotation['short_answers']:
                for sa in annotation['short_answers']:
                    start_ind = sa[f'start_{offset_unit}']
                    end_ind = sa[f'end_{offset_unit}']
                    ans = context[start_ind:end_ind]
                    if self.use_html:
                        short_answers.append(ans.decode())
                    else:
                        short_answers.append(' '.join(ans))
            elif (
                'yes_no_answer' in annotation
                and annotation['yes_no_answer']
                and annotation['yes_no_answer'] != 'NONE'
            ):
                short_answers.append(annotation['yes_no_answer'])
        return short_answers

    def load_from_chunk(self, chunk_idx: int):
        """
        Loads from a chunk of the dataset, given the chunk index.

        Returns a list of dictionaries. Each dictionary is an example from the
            main dataset and stores the components of that examples (e.g.,
            contenxt, question, candidate answers etc.) as key-value pairs.

        :param chunk_idx: the index of the chunk dataset chunk file.
        """

        def _extract_labels_indices(example, candidate_labels):
            labels = []
            for label in example['annotations']:
                label_ind = label['long_answer']['candidate_index']
                labels.append(candidate_labels[label_ind])
            return labels

        fname = f'nq-{self.dtype}-{str(chunk_idx).zfill(2)}.jsonl'
        fpath = os.path.join(self.dpath, fname)
        output = []
        with jsonlines.open(fpath, 'r') as fi:
            for example in fi:
                example_components = dict()
                example = self._simplify(example)
                question = example['question_text']
                if self.use_context:
                    context = example[_context_type_key(self.use_html)]
                    example_components['text'] = f'{context}\n{question}?'
                else:
                    example_components['text'] = f'{question}?'

                if self.use_long_answer:
                    example_components[
                        'long_answers_candidate'
                    ] = self._get_candidate_long_answers(example)
                    example_components['long_answers'] = _extract_labels_indices(
                        example, example_components['long_answers_candidate']
                    )
                else:
                    example_components['short_answers'] = self._get_short_answers(
                        example
                    )
                output.append(example_components)
        return output

    def create_message(self, example_components, entry_idx=0):
        label_key = 'long_answers' if self.use_long_answer else 'short_answers'
        message_dict = {
            'id': self.id,
            'text': example_components['text'],
            'labels': example_components[label_key] or [''],
            'episode_done': True,
        }
        if self.use_long_answer:
            message_dict['label_candidates'] = example_components[
                'long_answers_candidate'
            ]
        return message_dict


class DefaultTeacher(NaturalQuestionsTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import errno
import json
import logging
import threading
import time
import websocket
import parlai.chat_service.utils.logging as log_utils

SOCKET_TIMEOUT = 6


# Socket handler
class ChatServiceMessageSocket:
    """
    ChatServiceMessageSocket is a wrapper around websocket to forward messages from the
    remote server to the ChatServiceManager.
    """

    def __init__(self, server_url, port, message_callback):
        """
        server_url:           url at which the server is to be run
        port:                 port for the socket to operate on
        message_callback:     function to be called on incoming message objects (format: message_callback(self, data))
        """
        self.server_url = server_url
        self.port = port
        self.message_callback = message_callback

        self.ws = None
        self.last_pong = None
        self.alive = False

        # initialize the state
        self.listen_thread = None

        # setup the socket
        self.keep_running = True
        self._setup_socket()

    def _safe_send(self, data, force=False):
        if not self.alive and not force:
            # Try to wait a second to send a packet
            timeout = 1
            while timeout > 0 and not self.alive:
                time.sleep(0.1)
                timeout -= 0.1
            if not self.alive:
                # don't try to send a packet if we're still dead
                return False
        try:
            self.ws.send(data)
        except websocket.WebSocketConnectionClosedException:
            # The channel died mid-send, wait for it to come back up
            return False
        return True

    def _ensure_closed(self):
        try:
            self.ws.close()
        except websocket.WebSocketConnectionClosedException:
            pass

    def _send_world_alive(self):
        """
        Registers world with the passthrough server.
        """
        self._safe_send(
            json.dumps(
                {
                    'type': 'world_alive',
                    'content': {'id': 'WORLD_ALIVE', 'sender_id': 'world'},
                }
            ),
            force=True,
        )

    def _setup_socket(self):
        """
        Create socket handlers and registers the socket.
        """

        def on_socket_open(*args):
            log_utils.print_and_log(logging.DEBUG, 'Socket open: {}'.format(args))
            self._send_world_alive()

        def on_error(ws, error):
            try:
                if error.errno == errno.ECONNREFUSED:
                    self._ensure_closed()
                    self.use_socket = False
                    raise Exception("Socket refused connection, cancelling")
                else:
                    log_utils.print_and_log(
                        logging.WARN, 'Socket logged error: {}'.format(repr(error))
                    )
            except BaseException:
                if type(error) is websocket.WebSocketConnectionClosedException:
                    return  # Connection closed is noop
                log_utils.print_and_log(
                    logging.WARN,
                    'Socket logged error: {} Restarting'.format(repr(error)),
                )
                self._ensure_closed()

        def on_disconnect(*args):
            """
            Disconnect event is a no-op for us, as the server reconnects automatically
            on a retry.
            """
            log_utils.print_and_log(
                logging.INFO, 'World server disconnected: {}'.format(args)
            )
            self.alive = False
            self._ensure_closed()

        def on_message(*args):
            """
            Incoming message handler for messages from the FB user.
            """
            packet_dict = json.loads(args[1])
            if packet_dict['type'] == 'conn_success':
                self.alive = True
                return  # No action for successful connection
            if packet_dict['type'] == 'pong':
                self.last_pong = time.time()
                return  # No further action for pongs
            message_data = packet_dict['content']
            log_utils.print_and_log(
                logging.DEBUG, 'Message data received: {}'.format(message_data)
            )
            for message_packet in message_data['entry']:
                for message in message_packet['messaging']:
                    self.message_callback(message)

        def run_socket(*args):
            url_base_name = self.server_url.split('https://')[1]
            while self.keep_running:
                try:
                    sock_addr = "wss://{}/".format(url_base_name)
                    self.ws = websocket.WebSocketApp(
                        sock_addr,
                        on_message=on_message,
                        on_error=on_error,
                        on_close=on_disconnect,
                    )
                    self.ws.on_open = on_socket_open
                    self.ws.run_forever(ping_interval=1, ping_timeout=0.9)
                except Exception as e:
                    log_utils.print_and_log(
                        logging.WARN,
                        'Socket error {}, attempting restart'.format(repr(e)),
                    )
                time.sleep(0.2)

        # Start listening thread
        self.listen_thread = threading.Thread(
            target=run_socket, name='Main-Socket-Thread'
        )
        self.listen_thread.daemon = True
        self.listen_thread.start()
        time.sleep(1.2)
        while not self.alive:
            try:
                self._send_world_alive()
            except Exception:
                pass
            time.sleep(0.8)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Code for distilling a transformer/generator model.
"""

from typing import Optional
from parlai.core.params import ParlaiParser
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple, Type, Union

import torch
from torch import nn as nn
from torch.nn import functional as F

from parlai.agents.bart.bart import BartAgent
from parlai.agents.transformer.modules import (
    MultiHeadAttention,
    TransformerDecoderLayer,
    TransformerEncoderLayer,
)
from parlai.agents.transformer.transformer import TransformerGeneratorAgent
from parlai.core.agents import create_agent_from_model_file
from parlai.core.metrics import AverageMetric
from parlai.core.opt import Opt
from parlai.core.torch_agent import Batch
from parlai.core.torch_generator_agent import PPLMetric, TorchGeneratorAgent
from parlai.utils.misc import AttrDict
from parlai.utils.torch import NEAR_INF_FP16
from parlai.utils.typing import TShared


class OutputRecorder:
    """
    Saves all outputs from modules that it is registered to.
    """

    def __init__(self):
        self.outputs = []

    def __call__(self, module: nn.Module, module_in: Any, module_out: Any):
        self.outputs.append(module_out)

    def clear(self):
        self.outputs = []


class ForwardPassOutputs(AttrDict):
    """
    Provides dot access to all outputs of the forward passes of the encoder and decoder.
    """

    mask: torch.BoolTensor
    decoder_mask: torch.BoolTensor
    tokens_per_example: torch.Tensor
    num_tokens: torch.Tensor
    context_mask: torch.BoolTensor
    context_tokens_per_example: torch.Tensor
    num_context_tokens: torch.Tensor
    task_loss: torch.Tensor
    teacher_scores: torch.Tensor
    teacher_enc_output: torch.Tensor
    teacher_embedding_outputs: Dict[str, torch.Tensor]
    teacher_hidden_states: Dict[str, List[torch.Tensor]]
    teacher_attention_matrices: Dict[str, List[Dict[str, torch.Tensor]]]
    student_output: tuple
    student_scores: torch.Tensor
    student_enc_output: torch.Tensor
    student_embedding_outputs: Dict[str, torch.Tensor]
    student_hidden_states: Dict[str, List[torch.Tensor]]
    student_attention_matrices: Dict[str, List[Dict[str, torch.Tensor]]]

    def __init__(
        self,
        mask,
        decoder_mask,
        tokens_per_example,
        num_tokens,
        context_mask,
        context_tokens_per_example,
        num_context_tokens,
        task_loss,
        teacher_scores,
        teacher_enc_output,
        teacher_embedding_outputs,
        teacher_hidden_states,
        teacher_attention_matrices,
        student_output,
        student_scores,
        student_enc_output,
        student_embedding_outputs,
        student_hidden_states,
        student_attention_matrices,
    ):
        super().__init__(
            mask=mask,
            decoder_mask=decoder_mask,
            tokens_per_example=tokens_per_example,
            num_tokens=num_tokens,
            context_mask=context_mask,
            context_tokens_per_example=context_tokens_per_example,
            num_context_tokens=num_context_tokens,
            task_loss=task_loss,
            teacher_scores=teacher_scores,
            teacher_enc_output=teacher_enc_output,
            teacher_embedding_outputs=teacher_embedding_outputs,
            teacher_hidden_states=teacher_hidden_states,
            teacher_attention_matrices=teacher_attention_matrices,
            student_output=student_output,
            student_scores=student_scores,
            student_enc_output=student_enc_output,
            student_embedding_outputs=student_embedding_outputs,
            student_hidden_states=student_hidden_states,
            student_attention_matrices=student_attention_matrices,
        )


class AbstractDistillTransformerAgentMixin(ABC):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        agent = parser.add_argument_group('AbstractDistillTransformer arguments')
        agent.add_argument('--teacher-model', help='The teacher model file')
        agent.add_argument(
            '--task-loss-coeff',
            type=float,
            default=1,
            help='Coefficient on MLE loss on task',
        )
        agent.add_argument(
            '--encoder-loss-coeff',
            type=float,
            default=0,
            help='Coefficient on teacher loss on encoder output',
        )
        agent.add_argument(
            '--hidden-loss-coeff',
            type=float,
            default=0,
            help='Coefficient on teacher loss on encoder/decoder hidden layers',
        )
        agent.add_argument(
            '--pred-loss-coeff',
            type=float,
            default=0,
            help='Coefficient on KL teacher loss on prediction layer',
        )
        return agent

    def __init__(self, opt, shared=None):

        # Define coefficients
        self.task_loss_coeff = opt['task_loss_coeff']
        self.encoder_loss_coeff = opt['encoder_loss_coeff']
        self.hidden_loss_coeff = opt['hidden_loss_coeff']
        self.pred_loss_coeff = opt['pred_loss_coeff']

        assert (
            opt.get('model_parallel', False) is False
        ), 'model_parallel is not currently supported for distillation!'

        # Create teacher model
        if shared is None:
            to_copy = {'no_cuda', 'model_parallel', 'fp16', 'fp16_impl', 'gpu'}
            override = {k: opt[k] for k in to_copy}
            override['datatype'] = 'train:evalmode'  # Don't initialize the optimizer
            teacher_agent = create_agent_from_model_file(opt['teacher_model'], override)
            self.teacher_agent_opt = teacher_agent.opt
            self.teacher_model = teacher_agent.model
            self.teacher_model.eval()

        super().__init__(opt, shared)

    def build_model(self):

        student_model = super().build_model()

        teacher_model = self._get_teacher_model()

        # Define the mapping between teacher and student encoder layers
        self.student_num_enc_layers = len(student_model.encoder.layers)
        self.teacher_num_enc_layers = len(teacher_model.encoder.layers)
        self.enc_layer_ratio = self.teacher_num_enc_layers / self.student_num_enc_layers
        self.mapped_enc_layers = []
        for i in range(self.student_num_enc_layers):
            j = int((i + 1) * self.enc_layer_ratio) - 1
            self.mapped_enc_layers.append(j)

        # Define the mapping between teacher and student decoder layers
        self.student_num_dec_layers = len(student_model.decoder.layers)
        self.teacher_num_dec_layers = len(teacher_model.decoder.layers)
        self.dec_layer_ratio = self.teacher_num_dec_layers / self.student_num_dec_layers
        self.mapped_dec_layers = []
        for i in range(self.student_num_dec_layers):
            j = int((i + 1) * self.dec_layer_ratio) - 1
            self.mapped_dec_layers.append(j)

        # Register hooks to record outputs
        encoder_module_map = {
            'layers': TransformerEncoderLayer,
            'attentions': MultiHeadAttention,
        }
        decoder_module_map = {
            'layers': TransformerDecoderLayer,
            'attentions': MultiHeadAttention,
        }
        self.hooks = {
            'teacher': {
                'encoder': self._register_series_of_hooks(
                    model=teacher_model.encoder, module_map=encoder_module_map
                ),
                'decoder': self._register_series_of_hooks(
                    model=teacher_model.decoder, module_map=decoder_module_map
                ),
            },
            'student': {
                'encoder': self._register_series_of_hooks(
                    model=student_model.encoder, module_map=encoder_module_map
                ),
                'decoder': self._register_series_of_hooks(
                    model=student_model.decoder, module_map=decoder_module_map
                ),
            },
        }

        # Separately register hooks for the token embeddings, which are the same for
        # the encoder and decoder
        self.hooks['teacher']['embeddings'] = OutputRecorder()
        teacher_model.embeddings.register_forward_hook(
            self.hooks['teacher']['embeddings']
        )
        self.hooks['student']['embeddings'] = OutputRecorder()
        student_model.embeddings.register_forward_hook(
            self.hooks['student']['embeddings']
        )

        return student_model

    def _get_teacher_model(self) -> nn.Module:
        """
        Return the teacher model.

        This logic is needed because the teacher model may be wrapped by
        torch.nn.parallel.DistributedDataParallel.
        """
        if hasattr(self.teacher_model, 'module'):
            return self.teacher_model.module
        else:
            return self.teacher_model

    def _register_series_of_hooks(
        self, model: nn.Module, module_map: Dict[str, Type[nn.Module]]
    ) -> Dict[str, OutputRecorder]:
        """
        Register hooks in modules of the model, given the mapping of module types.

        `module_map` is a dict whose keys are module-type names and whose values are
        module types. For each module type, during each forward pass of `model`, all
        outputs of modules of that type will be saved to `hooks[module_type].outputs`.
        """
        hooks = {}
        for module_name, module_type in module_map.items():
            hooks[module_name] = OutputRecorder()
            for module in model.modules():
                if isinstance(module, module_type):
                    module.register_forward_hook(hooks[module_name])
        return hooks

    @abstractmethod
    def compute_loss(self, batch, return_output=False):
        """
        Return the loss.

        This will likely call self._perform_forward_passes().
        """

    def _perform_forward_passes(self, batch: Batch) -> ForwardPassOutputs:
        """
        Perform forward passes through the student and teacher and pass back outputs.
        """

        assert isinstance(self, TorchGeneratorAgent)
        # Code relies on methods

        mask = batch.label_vec != self.NULL_IDX

        self._clear_hook_outputs(self.hooks)

        # Forward pass through teacher model
        with torch.no_grad():
            teacher_scores, teacher_preds, teacher_enc_states = self.teacher_model(
                *self._model_input(batch), ys=batch.label_vec
            )
            teacher_enc_output, context_mask = teacher_enc_states

        # Forward pass through student model
        task_loss, student_output = super().compute_loss(batch, return_output=True)
        student_scores, student_preds, student_enc_states = student_output
        student_enc_output, _ = student_enc_states

        # Compile all outputs given the hooks
        teacher_embedding_outputs = self._extract_embedding_outputs(
            hooks=self.hooks['teacher']
        )
        student_embedding_outputs = self._extract_embedding_outputs(
            hooks=self.hooks['student']
        )
        teacher_hidden_states = self._extract_hidden_states(
            hooks=self.hooks['teacher'],
            num_enc_layers=self.teacher_num_enc_layers,
            num_dec_layers=self.teacher_num_dec_layers,
        )
        student_hidden_states = self._extract_hidden_states(
            hooks=self.hooks['student'],
            num_enc_layers=self.student_num_enc_layers,
            num_dec_layers=self.student_num_dec_layers,
        )
        teacher_attention_matrices = self._extract_attention_matrices(
            hooks=self.hooks['teacher'],
            num_enc_layers=self.teacher_num_enc_layers,
            num_dec_layers=self.teacher_num_dec_layers,
        )
        student_attention_matrices = self._extract_attention_matrices(
            hooks=self.hooks['student'],
            num_enc_layers=self.student_num_enc_layers,
            num_dec_layers=self.student_num_dec_layers,
        )
        self._clear_hook_outputs(self.hooks)

        tokens_per_example = mask.sum(dim=-1)  # Sum over tokens
        num_tokens = mask.sum()
        context_tokens_per_example = context_mask.sum(dim=-1)  # Sum over tokens
        num_context_tokens = context_mask.sum()

        # If needed, perform further manipulation of the mask tensor
        mask = self._manipulate_mask(
            mask=mask, student_scores=student_scores, batch=batch
        )
        decoder_mask = self._manipulate_mask(
            mask=mask, student_scores=student_embedding_outputs["decoder"], batch=batch
        )

        # Record teacher accuracy
        teacher_acc = ((student_preds == teacher_preds) * mask).sum(dim=-1)
        # Sum over tokens
        self.record_local_metric(
            'teacher_acc', AverageMetric.many(teacher_acc, tokens_per_example)
        )

        return ForwardPassOutputs(
            mask=mask,
            decoder_mask=decoder_mask,
            tokens_per_example=tokens_per_example,
            num_tokens=num_tokens,
            context_mask=context_mask,
            context_tokens_per_example=context_tokens_per_example,
            num_context_tokens=num_context_tokens,
            task_loss=task_loss,
            teacher_scores=teacher_scores,
            teacher_enc_output=teacher_enc_output,
            teacher_embedding_outputs=teacher_embedding_outputs,
            teacher_hidden_states=teacher_hidden_states,
            teacher_attention_matrices=teacher_attention_matrices,
            student_output=student_output,
            student_scores=student_scores,
            student_enc_output=student_enc_output,
            student_embedding_outputs=student_embedding_outputs,
            student_hidden_states=student_hidden_states,
            student_attention_matrices=student_attention_matrices,
        )

    def _manipulate_mask(
        self, mask: torch.BoolTensor, student_scores: torch.Tensor, batch: Batch
    ) -> torch.BoolTensor:
        """
        If necessary, perform further manipulations of the mask.

        Needed for BART-based student models to add in an extra start token.
        """
        if hasattr(super(), '_manipulate_mask'):
            # Defer to any agent-specific method for manipulating the mask
            return super()._manipulate_mask(
                mask=mask, student_scores=student_scores, batch=batch
            )
        else:
            return mask

    def _extract_embedding_outputs(
        self, hooks: Dict[str, Dict[str, OutputRecorder]]
    ) -> Dict[str, torch.Tensor]:
        """
        Extract out the encoder and decoder embedding outputs.
        """
        assert len(hooks['embeddings'].outputs) == 2
        return {
            'encoder': hooks['embeddings'].outputs[0],
            'decoder': hooks['embeddings'].outputs[1],
        }

    def _extract_hidden_states(
        self,
        hooks: Dict[str, Dict[str, OutputRecorder]],
        num_enc_layers: int,
        num_dec_layers: int,
    ) -> Dict[str, List[torch.Tensor]]:
        """
        Extract out encoder/decoder hidden states per layer.
        """
        assert len(hooks['encoder']['layers'].outputs) == num_enc_layers
        assert len(hooks['decoder']['layers'].outputs) == num_dec_layers
        return {
            'encoder': hooks['encoder']['layers'].outputs,
            'decoder': [out_[0] for out_ in hooks['decoder']['layers'].outputs],
        }

    def _extract_attention_matrices(
        self,
        hooks: Dict[str, Dict[str, OutputRecorder]],
        num_enc_layers: int,
        num_dec_layers: int,
    ) -> Dict[str, List[Dict[str, torch.Tensor]]]:
        """
        Extract out encoder/decoder attention matrices per layer and attention type.
        """
        assert len(hooks['encoder']['attentions'].outputs) == num_enc_layers
        assert len(hooks['decoder']['attentions'].outputs) == 2 * num_dec_layers
        output_idx = 2  # The position of the attention matrix among the outputs
        return {
            'encoder': [
                {
                    'self_attn': hooks['encoder']['attentions'].outputs[layer_idx][
                        output_idx
                    ]
                }
                for layer_idx in range(num_enc_layers)
            ],
            'decoder': [
                {
                    'self_attn': hooks['decoder']['attentions'].outputs[2 * layer_idx][
                        output_idx
                    ],
                    'encoder_attn': hooks['decoder']['attentions'].outputs[
                        2 * layer_idx + 1
                    ][output_idx],
                }
                for layer_idx in range(num_dec_layers)
            ],
        }

    def _clear_hook_outputs(self, hooks: Union[Dict[str, Any], OutputRecorder]):
        """
        Recursively clear outputs from all hooks.
        """
        if isinstance(hooks, dict):
            for subhooks in hooks.values():
                self._clear_hook_outputs(subhooks)
        else:
            # `hooks` is an OutputRecorder
            hooks.clear()

    def _get_encoder_loss(self, fwd_pass: ForwardPassOutputs) -> torch.Tensor:
        """
        Return the loss on the encoder's output layer.
        """
        assert isinstance(self, TorchGeneratorAgent)
        # Code relies on methods
        encoder_loss = F.mse_loss(
            input=fwd_pass.student_enc_output,
            target=fwd_pass.teacher_enc_output,
            reduction='none',
        )
        encoder_loss = encoder_loss.mean(dim=-1) * fwd_pass.context_mask
        # Avg over embedding dim
        self.record_local_metric(
            'enc_loss',
            AverageMetric.many(
                encoder_loss.sum(dim=-1), fwd_pass.context_tokens_per_example
            ),
        )  # Sum over token dim
        encoder_loss = encoder_loss.div(fwd_pass.num_context_tokens).sum()
        # Divide before summing over examples so that values don't get too large
        return encoder_loss

    def _get_embedding_losses(
        self, fwd_pass: ForwardPassOutputs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Return the encoder and decoder embedding losses.
        """
        assert isinstance(self, TorchGeneratorAgent)
        # Code relies on methods
        enc_emb_loss, enc_emb_loss_per_example = self._get_component_embedding_loss(
            student_emb_output=fwd_pass.student_embedding_outputs['encoder'],
            teacher_emb_output=fwd_pass.teacher_embedding_outputs['encoder'],
            mask=fwd_pass.context_mask,
            num_tokens=fwd_pass.num_context_tokens,
        )
        self.record_local_metric(
            'enc_emb_loss',
            AverageMetric.many(
                enc_emb_loss_per_example, fwd_pass.context_tokens_per_example
            ),
        )
        dec_emb_loss, dec_emb_loss_per_example = self._get_component_embedding_loss(
            student_emb_output=fwd_pass.student_embedding_outputs['decoder'],
            teacher_emb_output=fwd_pass.teacher_embedding_outputs['decoder'],
            mask=fwd_pass.decoder_mask,
            num_tokens=fwd_pass.num_tokens,
        )
        self.record_local_metric(
            'dec_emb_loss',
            AverageMetric.many(dec_emb_loss_per_example, fwd_pass.tokens_per_example),
        )
        return enc_emb_loss, dec_emb_loss

    def _get_component_embedding_loss(
        self,
        student_emb_output: torch.Tensor,
        teacher_emb_output: torch.Tensor,
        mask: torch.BoolTensor,
        num_tokens: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Compute the embedding loss for either the encoder or the decoder.
        """
        assert isinstance(self, TorchGeneratorAgent)
        # Code relies on methods
        raw_loss = F.mse_loss(
            input=student_emb_output, target=teacher_emb_output, reduction='none'
        )
        clamped_loss = torch.clamp(raw_loss, min=0, max=NEAR_INF_FP16)
        # Prevent infs from appearing in the loss term. Especially important with fp16
        masked_loss = clamped_loss.mean(dim=-1) * mask
        # Average over embedding dim
        embedding_loss_per_example = masked_loss.sum(dim=-1)  # Sum over token dim
        embedding_loss = masked_loss.div(num_tokens).sum()
        # Divide before summing over examples so that values don't get too large
        return embedding_loss, embedding_loss_per_example

    def _get_hidden_losses(
        self, fwd_pass: ForwardPassOutputs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Return the encoder and decoder hidden losses.
        """
        assert isinstance(self, TorchGeneratorAgent)
        # Code relies on methods
        enc_hidden_loss, enc_hidden_loss_per_example = self._get_component_hidden_loss(
            student_hidden_states=fwd_pass.student_hidden_states['encoder'],
            teacher_hidden_states=fwd_pass.teacher_hidden_states['encoder'],
            mask=fwd_pass.context_mask,
            num_tokens=fwd_pass.num_context_tokens,
            mapped_layers=self.mapped_enc_layers,
        )
        self.record_local_metric(
            'enc_hid_loss',
            AverageMetric.many(
                enc_hidden_loss_per_example, fwd_pass.context_tokens_per_example
            ),
        )
        dec_hidden_loss, dec_hidden_loss_per_example = self._get_component_hidden_loss(
            student_hidden_states=fwd_pass.student_hidden_states['decoder'],
            teacher_hidden_states=fwd_pass.teacher_hidden_states['decoder'],
            mask=fwd_pass.decoder_mask,
            num_tokens=fwd_pass.num_tokens,
            mapped_layers=self.mapped_dec_layers,
        )
        self.record_local_metric(
            'dec_hid_loss',
            AverageMetric.many(
                dec_hidden_loss_per_example, fwd_pass.tokens_per_example
            ),
        )
        return enc_hidden_loss, dec_hidden_loss

    def _get_component_hidden_loss(
        self,
        student_hidden_states: List[torch.Tensor],
        teacher_hidden_states: List[torch.Tensor],
        mask: torch.BoolTensor,
        num_tokens: torch.Tensor,
        mapped_layers: List[int],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Compute the loss across all hidden layers for either the encoder or the decoder.

        (The loss is averaged across all hidden layers and over the embedding dimension
        so that it doesn't get too high for fp16 tensors.)
        """
        per_layer_losses = []
        per_layer_per_example_losses = []
        for student_layer_idx, teacher_layer_idx in enumerate(mapped_layers):
            raw_layer_loss = F.mse_loss(
                input=student_hidden_states[student_layer_idx],
                target=teacher_hidden_states[teacher_layer_idx],
                reduction='none',
            )
            clamped_layer_loss = torch.clamp(raw_layer_loss, min=0, max=NEAR_INF_FP16)
            # Prevent infs from appearing in the loss term. Especially important with
            # fp16
            masked_layer_loss = clamped_layer_loss.mean(dim=-1) * mask
            # Average over embedding dim
            layer_loss_per_example = masked_layer_loss.sum(dim=-1)  # Sum over token dim
            layer_loss = masked_layer_loss.div(num_tokens).sum()
            # Divide before summing over examples so that values don't get too large
            per_layer_losses.append(layer_loss)
            per_layer_per_example_losses.append(layer_loss_per_example)
        hidden_loss = torch.stack(per_layer_losses).mean()
        hidden_loss_per_example = torch.stack(per_layer_per_example_losses, dim=1).mean(
            dim=1
        )
        return hidden_loss, hidden_loss_per_example

    def _get_attention_losses(
        self, fwd_pass: ForwardPassOutputs
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Return attention losses.

        Compute and return losses on encoder and decoder self-attention and decoder
        enc/dec attention.
        """
        enc_self_attn_loss = self._get_and_record_component_attention_loss(
            student_attention_matrices=fwd_pass.student_attention_matrices['encoder'],
            teacher_attention_matrices=fwd_pass.teacher_attention_matrices['encoder'],
            mask=fwd_pass.context_mask,
            tokens_per_example=fwd_pass.context_tokens_per_example,
            num_tokens=fwd_pass.num_context_tokens,
            mapped_layers=self.mapped_enc_layers,
            attn_type='self_attn',
            metric_name='enc_self_attn_loss',
        )
        dec_self_attn_loss = self._get_and_record_component_attention_loss(
            student_attention_matrices=fwd_pass.student_attention_matrices['decoder'],
            teacher_attention_matrices=fwd_pass.teacher_attention_matrices['decoder'],
            mask=fwd_pass.decoder_mask,
            tokens_per_example=fwd_pass.tokens_per_example,
            num_tokens=fwd_pass.num_tokens,
            mapped_layers=self.mapped_dec_layers,
            attn_type='self_attn',
            metric_name='dec_self_attn_loss',
        )
        enc_dec_attn_loss = self._get_and_record_component_attention_loss(
            student_attention_matrices=fwd_pass.student_attention_matrices['decoder'],
            teacher_attention_matrices=fwd_pass.teacher_attention_matrices['decoder'],
            mask=fwd_pass.decoder_mask,
            tokens_per_example=fwd_pass.tokens_per_example,
            num_tokens=fwd_pass.num_tokens,
            mapped_layers=self.mapped_dec_layers,
            attn_type='encoder_attn',
            metric_name='enc_dec_attn_loss',
        )
        return enc_self_attn_loss, dec_self_attn_loss, enc_dec_attn_loss

    def _get_and_record_component_attention_loss(
        self,
        teacher_attention_matrices: List[Dict[str, torch.Tensor]],
        student_attention_matrices: List[Dict[str, torch.Tensor]],
        mask: torch.BoolTensor,
        tokens_per_example: torch.Tensor,
        num_tokens: torch.Tensor,
        mapped_layers: List[int],
        attn_type: str,
        metric_name: str,
    ) -> torch.Tensor:
        """
        Calculate the given attention loss and register it as the given metric name.
        """

        assert isinstance(self, TorchGeneratorAgent)
        # Code relies on methods

        # Select the right attention matrices
        selected_student_attn_matrices = [
            layer_matrices[attn_type] for layer_matrices in student_attention_matrices
        ]
        selected_teacher_attn_matrices = [
            layer_matrices[attn_type] for layer_matrices in teacher_attention_matrices
        ]

        batch_size = mask.size(0)
        per_layer_losses = []
        per_layer_per_example_losses = []
        for student_layer_idx, teacher_layer_idx in enumerate(mapped_layers):
            raw_layer_loss = F.mse_loss(
                input=selected_student_attn_matrices[student_layer_idx],
                target=selected_teacher_attn_matrices[teacher_layer_idx],
                reduction='none',
            )
            clamped_layer_loss = torch.clamp(raw_layer_loss, min=0, max=NEAR_INF_FP16)
            # Prevent infs from appearing in the loss term. Especially important with
            # fp16
            reshaped_layer_loss = clamped_layer_loss.view(
                batch_size, -1, clamped_layer_loss.size(-2), clamped_layer_loss.size(-1)
            )
            # [batch size, n heads, query length, key length]
            mean_layer_loss = reshaped_layer_loss.mean(dim=(1, 3))
            # Take the mean over the attention heads and the key length
            assert mean_layer_loss.shape == mask.shape
            masked_layer_loss = mean_layer_loss * mask
            layer_loss_per_example = masked_layer_loss.sum(dim=-1)  # Sum over token dim
            layer_loss = masked_layer_loss.div(num_tokens).sum()
            # Divide before summing over examples so that values don't get too large
            per_layer_losses.append(layer_loss)
            per_layer_per_example_losses.append(layer_loss_per_example)
        attn_loss = torch.stack(per_layer_losses).mean()
        attn_loss_per_example = torch.stack(per_layer_per_example_losses, dim=1).mean(
            dim=1
        )

        # Record metric
        self.record_local_metric(
            metric_name, AverageMetric.many(attn_loss_per_example, tokens_per_example)
        )

        return attn_loss

    def _get_prediction_loss(self, fwd_pass: ForwardPassOutputs) -> torch.Tensor:
        """
        Calculate and return the KL loss on the teacher's prediction layer.

        Also record prediction-loss metrics.
        """
        assert isinstance(self, TorchGeneratorAgent)
        # Code relies on methods
        pred_loss = F.kl_div(
            F.log_softmax(fwd_pass.student_scores, dim=-1, dtype=torch.float),
            F.softmax(fwd_pass.teacher_scores, dim=-1, dtype=torch.float),
            reduction='none',
        ).type_as(fwd_pass.student_scores)
        pred_loss = pred_loss.sum(dim=-1) * fwd_pass.mask
        # Sum over dictionary
        self.record_local_metric(
            'pred_ppl',
            PPLMetric.many(pred_loss.sum(dim=-1), fwd_pass.tokens_per_example),
        )  # Sum over tokens
        self.record_local_metric(
            'pred_loss',
            AverageMetric.many(pred_loss.sum(dim=-1), fwd_pass.tokens_per_example),
        )  # Sum over tokens
        pred_loss = pred_loss.sum() / fwd_pass.num_tokens
        return pred_loss


class DistillTransformerAgentMixin(AbstractDistillTransformerAgentMixin):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group('DistillTransformer arguments')
        agent.add_argument(
            '--copy-teacher-weights',
            type='bool',
            default=True,
            help='Copy weights from the teacher model to the student model',
        )
        return agent

    def __init__(self, opt, shared=None):
        self.copy_teacher_weights = opt['copy_teacher_weights']
        if (
            opt.get('init_model')
            and os.path.isfile(opt['init_model'])
            and self.copy_teacher_weights
        ):
            raise Exception(
                "If --copy-teacher-weights is true, we cannot also copy over weights "
                "with --init-model!"
            )
        super().__init__(opt, shared)
        if shared is None:
            assert self.teacher_agent_opt['n_heads'] == opt['n_heads']

    def build_model(self):

        student_model = super().build_model()

        if self.copy_teacher_weights:

            teacher_model = self._get_teacher_model()

            # Initialize the embeddings
            student_model.encoder.embeddings.load_state_dict(
                teacher_model.encoder.embeddings.state_dict()
            )
            student_model.encoder.position_embeddings.load_state_dict(
                teacher_model.encoder.position_embeddings.state_dict()
            )
            student_model.decoder.embeddings.load_state_dict(
                teacher_model.decoder.embeddings.state_dict()
            )
            student_model.decoder.position_embeddings.load_state_dict(
                teacher_model.decoder.position_embeddings.state_dict()
            )

            # Initialize the encoder and decoder layers
            for student_idx, teacher_idx in enumerate(self.mapped_enc_layers):
                student_model.encoder.layers[student_idx].load_state_dict(
                    teacher_model.encoder.layers[teacher_idx].state_dict()
                )
            for student_idx, teacher_idx in enumerate(self.mapped_dec_layers):
                student_model.decoder.layers[student_idx].load_state_dict(
                    teacher_model.decoder.layers[teacher_idx].state_dict()
                )

        return student_model

    def compute_loss(self, batch, return_output=False):

        fwd_pass = self._perform_forward_passes(batch)

        # Calculate the loss on the encoder's output layer
        encoder_loss = self._get_encoder_loss(fwd_pass)

        # Calculate the loss on the encoder and decoder's hidden states
        enc_hidden_loss, dec_hidden_loss = self._get_hidden_losses(fwd_pass)

        # Calculate the KL loss on the teacher's prediction layer
        pred_loss = self._get_prediction_loss(fwd_pass)

        loss = (
            self.task_loss_coeff * fwd_pass.task_loss
            + self.encoder_loss_coeff * encoder_loss
            + self.hidden_loss_coeff * (enc_hidden_loss + dec_hidden_loss)
            + self.pred_loss_coeff * pred_loss
        )

        if return_output:
            return loss, fwd_pass.student_output
        else:
            return loss


class DistillNarrowTransformerAgentMixin(AbstractDistillTransformerAgentMixin):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group('DistillNarrowTransformer arguments')
        agent.add_argument(
            '--embedding-loss-coeff',
            type=float,
            default=0,
            help='Coefficient on teacher loss on embedding-layer output',
        )
        agent.add_argument(
            '--self-attn-loss-coeff',
            type=float,
            default=0,
            help='Coefficient on teacher loss on self-attention matrices',
        )
        agent.add_argument(
            '--enc-dec-attn-loss-coeff',
            type=float,
            default=0,
            help='Coefficient on teacher loss on enc/dec attention matrices',
        )
        return agent

    def __init__(self, opt, shared=None):
        self.embedding_loss_coeff = opt['embedding_loss_coeff']
        self.self_attn_loss_coeff = opt['self_attn_loss_coeff']
        self.enc_dec_attn_loss_coeff = opt['enc_dec_attn_loss_coeff']
        super().__init__(opt, shared)
        if shared is None:
            assert self.teacher_agent_opt['n_heads'] == opt['n_heads'] or (
                self.self_attn_loss_coeff == 0 and self.enc_dec_attn_loss_coeff == 0
            ), 'The number of attention heads can only differ between the student and teacher models if both attention loss coefficients are 0!'

    def build_model(self):
        student_model = super().build_model()

        # Build projection layers from the student to teacher hidden dim
        student_model.encoder_proj_layer = self._get_projection_layer(student_model)
        student_model.embedding_proj_layers = nn.ModuleDict(
            {
                'encoder': self._get_projection_layer(student_model),
                'decoder': self._get_projection_layer(student_model),
            }
        )
        student_model.hidden_proj_layers = nn.ModuleDict(
            {
                'encoder': nn.ModuleList(
                    [
                        self._get_projection_layer(student_model)
                        for _ in student_model.encoder.layers
                    ]
                ),
                'decoder': nn.ModuleList(
                    [
                        self._get_projection_layer(student_model)
                        for _ in student_model.decoder.layers
                    ]
                ),
            }
        )

        return student_model

    def _get_projection_layer(self, student_model):
        """
        Return a projection layer from the student hidden dim to the teacher hidden dim.
        """

        teacher_model = self._get_teacher_model()

        student_hidden_dim = student_model.encoder.dim
        teacher_hidden_dim = teacher_model.encoder.dim
        assert (
            student_hidden_dim == student_model.decoder.dim
            and teacher_hidden_dim == teacher_model.decoder.dim
        )

        layer = nn.Linear(student_hidden_dim, teacher_hidden_dim)

        # From TinyBERT's BertPreTrainedModel.init_bert_weights() method at
        # https://github.com/huawei-noah/Pretrained-Language-Model/blob/master/TinyBERT/transformer/modeling.py#L628
        layer.weight.data.normal_(mean=0.0, std=0.02)
        layer.bias.data.zero_()

        return layer

    def compute_loss(self, batch, return_output=False):

        fwd_pass = self._perform_forward_passes(batch)

        # Access the student model, which may be wrapped by
        # `torch.nn.parallel.DistributedDataParallel`
        if hasattr(self.model, 'module'):
            student_model = self.model.module
        else:
            student_model = self.model

        # Calculate the loss on the encoder's output layer
        fwd_pass.student_enc_output = student_model.encoder_proj_layer(
            fwd_pass.student_enc_output
        )
        # Pass encoder output through the corresponding projection layer
        encoder_loss = self._get_encoder_loss(fwd_pass)

        # Calculate the loss on the embedding layers
        for module, embedding_output in fwd_pass.student_embedding_outputs.items():
            # Loop over the encoder and the decoder
            fwd_pass.student_embedding_outputs[
                module
            ] = student_model.embedding_proj_layers[module](embedding_output)
            # Pass embedding output through the corresponding projection layer
        enc_emb_loss, dec_emb_loss = self._get_embedding_losses(fwd_pass)

        # Calculate the loss on the encoder and decoder's hidden states
        for module, per_layer_states in fwd_pass.student_hidden_states.items():
            # Loop over the encoder and the decoder
            assert len(per_layer_states) == len(
                student_model.hidden_proj_layers[module]
            )
            for layer_idx, hidden_state in enumerate(per_layer_states):
                # Loop over Transformer layers
                fwd_pass.student_hidden_states[module][
                    layer_idx
                ] = student_model.hidden_proj_layers[module][layer_idx](hidden_state)
                # Pass hidden state through the corresponding projection layer
        enc_hidden_loss, dec_hidden_loss = self._get_hidden_losses(fwd_pass)

        # Calculate the losses on the attention matrices
        if self.self_attn_loss_coeff != 0 or self.enc_dec_attn_loss_coeff != 0:
            (
                enc_self_attn_loss,
                dec_self_attn_loss,
                enc_dec_attn_loss,
            ) = self._get_attention_losses(fwd_pass)
        else:
            # Skip calculating the losses and just set them to 0 because they do not
            # form part of the loss function. This is useful if the number of attention
            # heads is different between the student and teacher, because in that case
            # that the attention query-key matrices will be of different shape.
            enc_self_attn_loss = dec_self_attn_loss = enc_dec_attn_loss = 0

        # Calculate the KL loss on the teacher's prediction layer
        pred_loss = self._get_prediction_loss(fwd_pass)

        loss = (
            self.task_loss_coeff * fwd_pass.task_loss
            + self.encoder_loss_coeff * encoder_loss
            + self.embedding_loss_coeff * (enc_emb_loss + dec_emb_loss)
            + self.hidden_loss_coeff * (enc_hidden_loss + dec_hidden_loss)
            + self.self_attn_loss_coeff * (enc_self_attn_loss + dec_self_attn_loss)
            + self.enc_dec_attn_loss_coeff * (enc_dec_attn_loss)
            + self.pred_loss_coeff * pred_loss
        )

        if return_output:
            return loss, fwd_pass.student_output
        else:
            return loss


class DistillTransformerAgent(DistillTransformerAgentMixin, TransformerGeneratorAgent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        DistillTransformerAgentMixin.add_cmdline_args(parser, partial_opt=partial_opt)
        TransformerGeneratorAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class DistillNarrowTransformerAgent(
    DistillNarrowTransformerAgentMixin, TransformerGeneratorAgent
):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        DistillNarrowTransformerAgentMixin.add_cmdline_args(
            parser, partial_opt=partial_opt
        )
        TransformerGeneratorAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class BartLikeAgent(BartAgent):
    """
    Subclass of the BART agent that prevents loading model weights from bart_large.
    """

    def __init__(self, opt: Opt, shared: TShared = None):
        # Just skip BartAgent._initialize_bart(opt)
        super(BartAgent, self).__init__(opt, shared)

    def _manipulate_mask(
        self, mask: torch.BoolTensor, student_scores: torch.Tensor, batch: Batch
    ) -> torch.BoolTensor:
        """
        Add one extra (masked-out) token to the mask, for compatibility with BART.

        Only necessary when examining decoder outputs directly.
        """
        if student_scores.size(1) == batch.label_vec.size(1) + 1:
            mask = torch.cat([mask.new_zeros([mask.size(0), 1]), mask], dim=1)
        return mask


class DistillBartAgent(DistillTransformerAgentMixin, BartLikeAgent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        DistillTransformerAgentMixin.add_cmdline_args(parser, partial_opt=partial_opt)
        BartLikeAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class DistillNarrowBartAgent(DistillNarrowTransformerAgentMixin, BartLikeAgent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        DistillNarrowTransformerAgentMixin.add_cmdline_args(
            parser, partial_opt=partial_opt
        )
        BartLikeAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import copy
import json
import os

from parlai.utils.io import PathManager
from parlai.core.teachers import DialogTeacher, MultiTaskTeacher
from parlai.tasks.multinli.agents import (
    convert_to_dialogData,
    BICLASS_LABELS,
    MULTINLI_LABELS,
)

from .build import build

ANLI = 'ANLI'
ANLI_PREFIX = 'anli_'
ANLI_VERSION = 'v0.1'
ANLI_LABEL_DICT = {'e': 'entailment', 'c': 'contradiction', 'n': 'neutral'}
ANLI_PREMISE_KEY = 'context'
ANLI_HYPO_KEY = 'hypothesis'
ANLI_ANSWER_KEY = 'label'
ANLI_ROUNDS = ['R1', 'R2', 'R3']


def _path(opt, anli_round):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if dt == 'train':
        suffix = 'train'
    elif dt == 'valid':
        suffix = 'dev'
    elif dt == 'test':
        suffix = 'test'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(
        opt['datapath'], ANLI, ANLI_PREFIX + ANLI_VERSION, anli_round, suffix + '.jsonl'
    )

    return data_path


class RoundBaseTeacher(DialogTeacher):
    """
    Base class for teachers in all 3 rounds in ANLI tasks.

    ``RoundBaseTeacher`` derives anli_round (the round index of ANLI task which consists of 3 rounds NLI tasks with
    increasing difficulty. (See https://arxiv.org/abs/1910.14599 for more information)  from ``opt['task']``.
    ``anli_round`` is used to set the correct path to the downloaded data file for thaue specified round.

    ``RoundBaseTeacher`` also parses the requested dialog format(text format w/ or w/o special tokens 'Hypothesis',
    'Premise') and number of classes (3 classes entailment, neutral, contradiction or 2 classes contradiction, not
    contradiction) desired for particular training purposes.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = parser.add_argument_group('RoundBase Teacher Args')
        parser.add_argument(
            '-dfm',
            '--dialog-format',
            type='bool',
            default=False,
            help="True if one would like to convert to a dialogue format without special tokens such as 'Premise'"
            " and 'Hypothesis' (default: False).",
        )
        parser.add_argument(
            '-bcl',
            '--binary-classes',
            type='bool',
            default=False,
            help="True if label candidates are (contradiction, not_contradiction), and (entailment, contradiction, "
            "neutral) otherwise (default: False).",
        )
        return parser

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        self.anli_round = (
            opt['task'].split(':')[1] if len(opt['task'].split(':')) > 1 else None
        )
        self.anli_round = self.anli_round.upper()
        if self.anli_round not in ANLI_ROUNDS:
            raise KeyError(f"Undefined task round: {self.anli_round}.")

        data_path = _path(opt, self.anli_round)
        opt['datafile'] = data_path
        self.dialog_format = opt.get('dialog_format', False)
        self.binary_classes = opt.get('binary_classes', False)
        self.id = opt['task'].upper()
        super().__init__(opt, shared)

    def label_candidates(self):
        if self.binary_classes:
            return BICLASS_LABELS
        return MULTINLI_LABELS

    def setup_data(self, path):
        print('loading: ' + path)
        with PathManager.open(path, 'r') as data_file:
            for pair_line in data_file:
                pair = json.loads(pair_line)
                if pair[ANLI_ANSWER_KEY] == '-':
                    continue

                label_raw = pair[ANLI_ANSWER_KEY]
                if label_raw in ANLI_LABEL_DICT:
                    label_raw = ANLI_LABEL_DICT[label_raw]

                question, answers, clas = convert_to_dialogData(
                    premise_raw=pair[ANLI_PREMISE_KEY],
                    hypo_raw=pair[ANLI_HYPO_KEY],
                    answer_raw=label_raw,
                    dialog_format=self.dialog_format,
                    binary_classes=self.binary_classes,
                )

                yield (question, answers, None, clas), True


class R1Teacher(RoundBaseTeacher):
    pass


class R2Teacher(RoundBaseTeacher):
    pass


class R3Teacher(RoundBaseTeacher):
    pass


class DefaultTeacher(MultiTaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = parser.add_argument_group('ANLI Teacher Args')
        parser.add_argument(
            '-dfm',
            '--dialog-format',
            type='bool',
            default=False,
            help="True if one would like to convert to a dialogue format without special tokens such as 'Premise'"
            " and 'Hypothesis' (default: False).",
        ),
        parser.add_argument(
            '-bcl',
            '--binary-classes',
            type='bool',
            default=False,
            help="True if label candidates are (contradiction, not_contradiction), and (entailment, contradiction, "
            "neutral) otherwise (default: False).",
        )
        return parser

    def __init__(self, opt, shared=None):
        anli_tasks = ['anli:r1', 'anli:r2', 'anli:r3']
        opt = copy.deepcopy(opt)
        opt['task'] = ','.join(anli_tasks)
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Websocket Runner.
"""
from parlai.core.params import ParlaiParser
from parlai.chat_service.services.websocket.websocket_manager import WebsocketManager
import parlai.chat_service.utils.config as config_utils


SERVICE_NAME = 'websocket'


def setup_args():
    """
    Set up args.
    """
    parser = ParlaiParser(False, False)
    parser.add_parlai_data_path()
    parser.add_websockets_args()
    return parser.parse_args()


def run(opt):
    """
    Run MessengerManager.
    """
    opt['service'] = SERVICE_NAME
    manager = WebsocketManager(opt)
    try:
        manager.start_task()
    except BaseException:
        raise
    finally:
        manager.shutdown()


if __name__ == '__main__':
    opt = setup_args()
    config_path = opt.get('config_path')
    config = config_utils.parse_configuration_file(config_path)
    opt.update(config['world_opt'])
    opt['config'] = config
    run(opt)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# hack to make sure -m transformer/generator works as expected
"""
Poly-encoder Agent.
"""

from parlai.core.params import ParlaiParser
from typing import Any, Dict, Optional, Tuple

import torch

from parlai.core.opt import Opt
from parlai.core.torch_ranker_agent import TorchRankerAgent
from parlai.utils.misc import recursive_getattr
from parlai.utils.logging import logging

from .biencoder import AddLabelFixedCandsTRA
from .modules import (
    BasicAttention,
    MultiHeadAttention,
    TransformerEncoder,
    get_n_positions_from_options,
)
from .transformer import TransformerRankerAgent


class PolyencoderAgent(TorchRankerAgent):
    """
    Poly-encoder Agent.

    Equivalent of bert_ranker/polyencoder and biencoder_multiple_output but does not
    rely on an external library (hugging face).
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        TransformerRankerAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group('Polyencoder Arguments')
        agent.add_argument(
            '--polyencoder-type',
            type=str,
            default='codes',
            choices=['codes', 'n_first'],
            help='Type of polyencoder, either we compute'
            'vectors using codes + attention, or we '
            'simply take the first N vectors.',
            recommended='codes',
        )
        agent.add_argument(
            '--poly-n-codes',
            type=int,
            default=64,
            help='number of vectors used to represent the context'
            'in the case of n_first, those are the number'
            'of vectors that are considered.',
            recommended=64,
        )
        agent.add_argument(
            '--poly-attention-type',
            type=str,
            default='basic',
            choices=['basic', 'sqrt', 'multihead'],
            help='Type of the top aggregation layer of the poly-'
            'encoder (where the candidate representation is'
            'the key)',
            recommended='basic',
        )
        agent.add_argument(
            '--poly-attention-num-heads',
            type=int,
            default=4,
            help='In case poly-attention-type is multihead, '
            'specify the number of heads',
        )

        # Those arguments are here in case where polyencoder type is 'code'
        agent.add_argument(
            '--codes-attention-type',
            type=str,
            default='basic',
            choices=['basic', 'sqrt', 'multihead'],
            help='Type ',
            recommended='basic',
        )
        agent.add_argument(
            '--codes-attention-num-heads',
            type=int,
            default=4,
            help='In case codes-attention-type is multihead, '
            'specify the number of heads',
        )
        return agent

    @classmethod
    def upgrade_opt(cls, opt_from_disk: Opt):
        # call the parent upgrades
        opt_from_disk = super(PolyencoderAgent, cls).upgrade_opt(opt_from_disk)

        polyencoder_attention_keys_value = opt_from_disk.get(
            'polyencoder_attention_keys'
        )
        if polyencoder_attention_keys_value is not None:
            # 2020-02-19 We are deprecating this flag because it was used for a one-time
            # set of experiments and won't be used again. This flag was defaulted to
            # 'context', so throw an exception otherwise.
            if polyencoder_attention_keys_value == 'context':
                del opt_from_disk['polyencoder_attention_keys']
            else:
                raise NotImplementedError(
                    'This --polyencoder-attention-keys mode (found in commit 06f0d9f) is no longer supported!'
                )

        return opt_from_disk

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.rank_loss = torch.nn.CrossEntropyLoss(reduce=True, size_average=True)
        if self.use_cuda:
            self.rank_loss.cuda()

    def build_model(self, states=None):
        """
        Return built model.
        """
        return PolyEncoderModule(self.opt, self.dict, self.NULL_IDX)

    def vectorize(self, *args, **kwargs):
        """
        Add the start and end token to the labels.
        """
        kwargs['add_start'] = True
        kwargs['add_end'] = True
        obs = super().vectorize(*args, **kwargs)
        return obs

    def _set_text_vec(self, *args, **kwargs):
        """
        Add the start and end token to the text.
        """
        obs = super()._set_text_vec(*args, **kwargs)
        if 'text_vec' in obs and 'added_start_end_tokens' not in obs:
            obs.force_set(
                'text_vec', self._add_start_end_tokens(obs['text_vec'], True, True)
            )
            obs['added_start_end_tokens'] = True
        return obs

    def vectorize_fixed_candidates(self, *args, **kwargs):
        """
        Vectorize fixed candidates.

        Override to add start and end token when computing the candidate encodings in
        interactive mode.
        """
        kwargs['add_start'] = True
        kwargs['add_end'] = True
        return super().vectorize_fixed_candidates(*args, **kwargs)

    def _make_candidate_encs(self, vecs):
        """
        Make candidate encs.

        The polyencoder module expects cand vecs to be 3D while torch_ranker_agent
        expects it to be 2D. This requires a little adjustment (used in interactive mode
        only)
        """
        rep = super()._make_candidate_encs(vecs)
        return rep.transpose(0, 1).contiguous()

    def encode_candidates(self, padded_cands):
        """
        Encode candidates.
        """
        padded_cands = padded_cands.unsqueeze(1)
        _, _, cand_rep = self.model(cand_tokens=padded_cands)
        return cand_rep

    def score_candidates(self, batch, cand_vecs, cand_encs=None):
        """
        Score candidates.

        The Poly-encoder encodes the candidate and context independently. Then, the
        model applies additional attention before ultimately scoring a candidate.
        """
        bsz = self._get_batch_size(batch)
        ctxt_rep, ctxt_rep_mask, _ = self.model(**self._model_context_input(batch))

        if cand_encs is not None:
            if bsz == 1:
                cand_rep = cand_encs
            else:
                cand_rep = cand_encs.expand(bsz, cand_encs.size(1), -1)
        # bsz x num cands x seq len
        elif len(cand_vecs.shape) == 3:
            _, _, cand_rep = self.model(cand_tokens=cand_vecs)
        # bsz x seq len (if batch cands) or num_cands x seq len (if fixed cands)
        elif len(cand_vecs.shape) == 2:
            _, _, cand_rep = self.model(cand_tokens=cand_vecs.unsqueeze(1))
            num_cands = cand_rep.size(0)  # will be bsz if using batch cands
            cand_rep = cand_rep.expand(num_cands, bsz, -1).transpose(0, 1).contiguous()
        scores = self.model(
            ctxt_rep=ctxt_rep, ctxt_rep_mask=ctxt_rep_mask, cand_rep=cand_rep
        )
        return scores

    def _get_batch_size(self, batch) -> int:
        """
        Return the size of the batch.

        Can be overridden by subclasses that do not always have text input.
        """
        return batch.text_vec.size(0)

    def _model_context_input(self, batch) -> Dict[str, Any]:
        """
        Create the input context value for the model.

        Must return a dictionary.  This will be passed directly into the model via
        `**kwargs`, i.e.,

        >>> model(**_model_context_input(batch))

        This is intentionally overridable so that richer models can pass additional
        inputs.
        """
        return {'ctxt_tokens': batch.text_vec}

    def load_state_dict(self, state_dict):
        """
        Override to account for codes.
        """
        if self.model.type == 'codes' and 'codes' not in state_dict:
            state_dict['codes'] = self.model.codes
        super().load_state_dict(state_dict)

    def _resize_token_embeddings(self, state_dict, msg=None):
        """
        Resize the token embeddings when adding extra special tokens.

        H/t TransformerGenerator._resize_token_embeddings for inspiration.
        """
        # map extra special tokens carefully
        new_size = self.model.encoder_ctxt.embeddings.weight.size()[0]
        orig_size = state_dict['encoder_ctxt.embeddings.weight'].size()[0]
        logging.info(f'Resizing token embeddings from {orig_size} to {new_size}')
        if new_size <= orig_size:
            # new size should be greater than original size,
            # as we are adding special tokens
            raise RuntimeError(msg)

        for emb_weights in [
            'encoder_ctxt.embeddings.weight',
            'encoder_cand.embeddings.weight',
        ]:
            # get new_embs
            old_embs = state_dict[emb_weights]
            new_embs = recursive_getattr(self.model, emb_weights).to(old_embs.device)
            # copy over old weights
            new_embs.data[:orig_size, :] = old_embs.data[:orig_size, :]
            # reset in state dict
            state_dict[emb_weights] = new_embs

        return state_dict


class PolyEncoderModule(torch.nn.Module):
    """
    Poly-encoder model.

    See https://arxiv.org/abs/1905.01969 for more details
    """

    def __init__(self, opt, dict_, null_idx):
        super(PolyEncoderModule, self).__init__()
        self.null_idx = null_idx
        self.encoder_ctxt = self.get_encoder(
            opt=opt,
            dict_=dict_,
            null_idx=null_idx,
            reduction_type=None,
            for_context=True,
        )
        self.encoder_cand = self.get_encoder(
            opt=opt,
            dict_=dict_,
            null_idx=null_idx,
            reduction_type=opt['reduction_type'],
            for_context=False,
        )

        self.type = opt['polyencoder_type']
        self.n_codes = opt['poly_n_codes']
        self.attention_type = opt['poly_attention_type']
        self.attention_num_heads = opt['poly_attention_num_heads']
        self.codes_attention_type = opt['codes_attention_type']
        self.codes_attention_num_heads = opt['codes_attention_num_heads']
        embed_dim = opt['embedding_size']

        # In case it's a polyencoder with code.
        if self.type == 'codes':
            # experimentally it seems that random with size = 1 was good.
            codes = torch.empty(self.n_codes, embed_dim)
            codes = torch.nn.init.uniform_(codes)
            self.codes = torch.nn.Parameter(codes)

            # The attention for the codes.
            if self.codes_attention_type == 'multihead':
                self.code_attention = MultiHeadAttention(
                    self.codes_attention_num_heads, embed_dim, opt['dropout']
                )
            elif self.codes_attention_type == 'sqrt':
                self.code_attention = PolyBasicAttention(
                    self.type, self.n_codes, dim=2, attn='sqrt', get_weights=False
                )
            elif self.codes_attention_type == 'basic':
                self.code_attention = PolyBasicAttention(
                    self.type, self.n_codes, dim=2, attn='basic', get_weights=False
                )

        # The final attention (the one that takes the candidate as key)
        if self.attention_type == 'multihead':
            self.attention = MultiHeadAttention(
                self.attention_num_heads, opt['embedding_size'], opt['dropout']
            )
        else:
            self.attention = PolyBasicAttention(
                self.type,
                self.n_codes,
                dim=2,
                attn=self.attention_type,
                get_weights=False,
            )

    def get_encoder(self, opt, dict_, null_idx, reduction_type, for_context: bool):
        """
        Return encoder, given options.

        :param opt:
            opt dict
        :param dict:
            dictionary agent
        :param null_idx:
            null/pad index into dict
        :param reduction_type:
            reduction type for the encoder
        :param for_context:
            whether this is the context encoder (as opposed to the candidate encoder).
            Useful for subclasses.
        :return:
            a TransformerEncoder, initialized correctly
        """
        n_positions = get_n_positions_from_options(opt)
        embeddings = self._get_embeddings(
            dict_=dict_, null_idx=null_idx, embedding_size=opt['embedding_size']
        )
        return TransformerEncoder(
            n_heads=opt['n_heads'],
            n_layers=opt['n_layers'],
            embedding_size=opt['embedding_size'],
            ffn_size=opt['ffn_size'],
            vocabulary_size=len(dict_),
            embedding=embeddings,
            dropout=opt['dropout'],
            attention_dropout=opt['attention_dropout'],
            relu_dropout=opt['relu_dropout'],
            padding_idx=null_idx,
            learn_positional_embeddings=opt['learn_positional_embeddings'],
            embeddings_scale=opt['embeddings_scale'],
            reduction_type=reduction_type,
            n_positions=n_positions,
            n_segments=opt.get('n_segments', 2),
            activation=opt['activation'],
            variant=opt['variant'],
            output_scaling=opt['output_scaling'],
        )

    def _get_embeddings(self, dict_, null_idx, embedding_size):
        embeddings = torch.nn.Embedding(
            len(dict_), embedding_size, padding_idx=null_idx
        )
        torch.nn.init.normal_(embeddings.weight, 0, embedding_size ** -0.5)
        return embeddings

    def attend(self, attention_layer, queries, keys, values, mask):
        """
        Apply attention.

        :param attention_layer:
            nn.Module attention layer to use for the attention
        :param queries:
            the queries for attention
        :param keys:
            the keys for attention
        :param values:
            the values for attention
        :param mask:
            mask for the attention keys

        :return:
            the result of applying attention to the values, with weights computed
            wrt to the queries and keys.
        """
        if keys is None:
            keys = values
        if isinstance(attention_layer, PolyBasicAttention):
            return attention_layer(queries, keys, mask_ys=mask, values=values)
        elif isinstance(attention_layer, MultiHeadAttention):
            return attention_layer(queries, keys, values, mask)[0]
        else:
            raise Exception('Unrecognized type of attention')

    def encode(
        self, cand_tokens: Optional[torch.Tensor], **ctxt_inputs: torch.Tensor
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]:
        """
        Encode a text sequence.

        :param ctxt_inputs:
            Dictionary of context inputs. If not empty, should contain at least
            'ctxt_tokens', a 2D long tensor of shape batchsize x sent_len
        :param cand_tokens:
            3D long tensor, batchsize x num_cands x sent_len
            Note this will actually view it as a 2D tensor
        :return:
            (ctxt_rep, ctxt_mask, cand_rep)
            - ctxt_rep 3D float tensor, batchsize x n_codes x dim
            - ctxt_mask byte:  batchsize x n_codes (all 1 in case
            of polyencoder with code. Which are the vectors to use
            in the ctxt_rep)
            - cand_rep (3D float tensor) batchsize x num_cands x dim
        """
        cand_embed = None
        ctxt_rep = None
        ctxt_rep_mask = None

        if cand_tokens is not None:
            assert len(cand_tokens.shape) == 3
            bsz = cand_tokens.size(0)
            num_cands = cand_tokens.size(1)
            cand_embed = self.encoder_cand(cand_tokens.view(bsz * num_cands, -1))
            cand_embed = cand_embed.view(bsz, num_cands, -1)

        if len(ctxt_inputs) > 0:
            assert 'ctxt_tokens' in ctxt_inputs
            if ctxt_inputs['ctxt_tokens'] is not None:
                assert len(ctxt_inputs['ctxt_tokens'].shape) == 2
            bsz = self._get_context_batch_size(**ctxt_inputs)
            # get context_representation. Now that depends on the cases.
            ctxt_out, ctxt_mask = self.encoder_ctxt(
                **self._context_encoder_input(ctxt_inputs)
            )
            dim = ctxt_out.size(2)

            if self.type == 'codes':
                ctxt_rep = self.attend(
                    self.code_attention,
                    queries=self.codes.repeat(bsz, 1, 1),
                    keys=ctxt_out,
                    values=ctxt_out,
                    mask=ctxt_mask,
                )
                ctxt_rep_mask = ctxt_rep.new_ones(bsz, self.n_codes).byte()

            elif self.type == 'n_first':
                # Expand the output if it is not long enough
                if ctxt_out.size(1) < self.n_codes:
                    difference = self.n_codes - ctxt_out.size(1)
                    extra_rep = ctxt_out.new_zeros(bsz, difference, dim)
                    ctxt_rep = torch.cat([ctxt_out, extra_rep], dim=1)
                    extra_mask = ctxt_mask.new_zeros(bsz, difference)
                    ctxt_rep_mask = torch.cat([ctxt_mask, extra_mask], dim=1)
                else:
                    ctxt_rep = ctxt_out[:, 0 : self.n_codes, :]
                    ctxt_rep_mask = ctxt_mask[:, 0 : self.n_codes]

        return ctxt_rep, ctxt_rep_mask, cand_embed

    def _get_context_batch_size(self, **ctxt_inputs: torch.Tensor) -> int:
        """
        Return the batch size of the context.

        Can be overridden by subclasses that do not always have text tokens in the
        context.
        """
        return ctxt_inputs['ctxt_tokens'].size(0)

    def _context_encoder_input(self, ctxt_inputs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Return the inputs to the context encoder as a dictionary.

        Must return a dictionary.  This will be passed directly into the model via
        `**kwargs`, i.e.,

        >>> encoder_ctxt(**_context_encoder_input(ctxt_inputs))

        This is needed because the context encoder's forward function may have different
        argument names than that of the model itself. This is intentionally overridable
        so that richer models can pass additional inputs.
        """
        assert set(ctxt_inputs.keys()) == {'ctxt_tokens'}
        return {'input': ctxt_inputs['ctxt_tokens']}

    def score(self, ctxt_rep, ctxt_rep_mask, cand_embed):
        """
        Score the candidates.

        :param ctxt_rep:
            3D float tensor, bsz x ctxt_len x dim
        :param ctxt_rep_mask:
            2D byte tensor, bsz x ctxt_len, in case there are some elements
            of the ctxt that we should not take into account.
        :param cand_embed: 3D float tensor, bsz x num_cands x dim

        :return: scores, 2D float tensor: bsz x num_cands
        """
        # reduces the context representation to a 3D tensor bsz x num_cands x dim
        ctxt_final_rep = self.attend(
            self.attention, cand_embed, ctxt_rep, ctxt_rep, ctxt_rep_mask
        )
        scores = torch.sum(ctxt_final_rep * cand_embed, 2)
        return scores

    def forward(
        self,
        cand_tokens=None,
        ctxt_rep=None,
        ctxt_rep_mask=None,
        cand_rep=None,
        **ctxt_inputs,
    ):
        """
        Forward pass of the model.

        Due to a limitation of parlai, we have to have one single model
        in the agent. And because we want to be able to use data-parallel,
        we need to have one single forward() method.
        Therefore the operation_type can be either 'encode' or 'score'.

        :param ctxt_inputs:
            Dictionary of context inputs. Will include at least 'ctxt_tokens',
            containing tokenized contexts
        :param cand_tokens:
            tokenized candidates
        :param ctxt_rep:
            (bsz x num_codes x hsz)
            encoded representation of the context. If self.type == 'codes', these
            are the context codes. Otherwise, they are the outputs from the
            encoder
        :param ctxt_rep_mask:
            mask for ctxt rep
        :param cand_rep:
            encoded representation of the candidates
        """
        if len(ctxt_inputs) > 0 or cand_tokens is not None:
            return self.encode(cand_tokens=cand_tokens, **ctxt_inputs)
        elif (
            ctxt_rep is not None and ctxt_rep_mask is not None and cand_rep is not None
        ):
            return self.score(ctxt_rep, ctxt_rep_mask, cand_rep)
        raise Exception('Unsupported operation')


class PolyBasicAttention(BasicAttention):
    """
    Override basic attention to account for edge case for polyencoder.
    """

    def __init__(self, poly_type, n_codes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.poly_type = poly_type
        self.n_codes = n_codes

    def forward(self, *args, **kwargs):
        """
        Forward pass.

        Account for accidental dimensionality reduction when num_codes is 1 and the
        polyencoder type is 'codes'
        """
        lhs_emb = super().forward(*args, **kwargs)
        if self.poly_type == 'codes' and self.n_codes == 1 and len(lhs_emb.shape) == 2:
            lhs_emb = lhs_emb.unsqueeze(self.dim - 1)
        return lhs_emb


class IRFriendlyPolyencoderAgent(AddLabelFixedCandsTRA, PolyencoderAgent):
    """
    Poly-encoder agent that allows for adding label to fixed cands.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add cmd line args.
        """
        AddLabelFixedCandsTRA.add_cmdline_args(parser, partial_opt=partial_opt)
        PolyencoderAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.convai2.agents import BothTeacher as OrigConvai2Teacher
from parlai.tasks.md_gender.build import build
from parlai.utils.misc import warn_once

import parlai.tasks.md_gender.utils as gend_utils

from copy import deepcopy
import json
import random
import os


gender_convert = {'man': 'male', 'woman': 'female', 'neutral': 'unknown', None: None}


def convert_text(text):
    """
    Normalize text.
    """
    new_text = text.lower()
    for x in gend_utils.CONTRACTION_SPACES:
        if x[1] in text:
            new_text = new_text.replace(x[1], x[0])
    for x in gend_utils.CONTRACTION_LEFT_SPACES:
        if x[1] in text:
            new_text = new_text.replace(x[1], x[0])
    for x in gend_utils.CONTRACTION_RIGHT_SPACES:
        if x[1] in text:
            new_text = new_text.replace(x[1], x[0])

    return new_text


class Convai2Teacher(FixedDialogTeacher):
    """
    Predict the gender of character given the dialogue utterance.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = gend_utils.add_common_args(parser)
        agent = parser.add_argument_group('ConvAI2 gender')
        agent.add_argument(
            '--convai2-use-probably',
            type='bool',
            default=True,
            help='Use the probable gender of the persona as '
            'determined by crowdworkers, instead of the '
            'definitive one',
        )
        agent.add_argument(
            '--labels-to-use',
            type=str,
            default='all',
            choices=['all', 'self', 'partner'],
            help='Which labels to use for this teacher',
        )
        return parser

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.fixed_random = random.Random(42)
        self.label_candidates = [gend_utils.MASC, gend_utils.FEM, gend_utils.NEUTRAL]
        self.use_probably = opt['convai2_use_probably']
        self.labels_to_use = opt['labels_to_use']
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = opt['add_unknown_classes'] and self.is_train

        if shared and 'data' in shared:
            self.data = shared['data']
            self.persona_map = shared['persona_map']
        else:
            self.missing_cnt = 0
            self._load_persona_map(opt)
            self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                exclude_lst = gend_utils.ABOUT_CANDS
                self.data = gend_utils.balance_data(
                    self.data, exclude_labels=exclude_lst
                )

        self.label_candidates = gend_utils.ALL_CANDS

        opt = deepcopy(opt)
        super().__init__(opt, shared)
        self.reset()

    def _load_persona_map(self, opt):
        build(opt)
        persona_map_path = os.path.join(
            opt['datapath'],
            'md_gender',
            'data_to_release',
            'convai2',
            'convai2_all_personas_map.json',
        )
        with open(persona_map_path, 'rb') as f:
            self.persona_map = json.load(f)

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def get_gender_annotations(self, lst):
        """
        Take a list of personas and find the gender.
        """
        new_lst = sorted([convert_text(persona) for persona in lst])
        persona_key = ' '.join(new_lst)
        if persona_key in self.persona_map:
            return self.persona_map[persona_key]
        else:
            return {'gender': 'neutral', 'probably': None, 'missing': True}

    def get_genders(self, your_persona, partner_persona):
        partner_persona_gender = self.get_gender_annotations(partner_persona)
        your_persona_gender = self.get_gender_annotations(your_persona)
        if 'missing' in partner_persona_gender:
            self.missing_cnt += 1
        if 'missing' in your_persona_gender:
            self.missing_cnt += 1

        partner_prob = None
        your_prob = None
        if self.use_probably:
            partner = partner_persona_gender['probably']
            if partner is None:
                partner = partner_persona_gender['gender']
            your = your_persona_gender['probably']
            if your is None:
                your = your_persona_gender['gender']
        else:
            partner = partner_persona_gender['gender']
            your = your_persona_gender['gender']
            partner_prob = partner_persona_gender['probably']
            your_prob = your_persona_gender['probably']

        your = gender_convert[your]
        your_prob = gender_convert[your_prob]
        partner = gender_convert[partner]
        partner_prob = gender_convert[partner_prob]

        return your, your_prob, partner, partner_prob

    def _setup_data(self, opt):
        counts = {
            'partner': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
            'self': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
        }

        dt = opt['datatype'].split(':')[0]
        if dt == 'test':
            warn_once('No test set; switching to valid')
            dt = 'valid'

        # build data
        print('[ Building data ... ]')
        new_eps = []
        orig_teacher = OrigConvai2Teacher(opt)
        total_exs = orig_teacher.num_examples()
        num_exs = 0
        while num_exs < total_exs:
            current_episode = []
            episode_done = False

            while not episode_done:
                # TODO: eventually all teachers should return Messages, so
                # we should assert this
                action = Message(orig_teacher.act())
                current_episode.append(action)
                episode_done = action.get('episode_done', False)
                num_exs += 1

            # now we have the entire episode,... do something
            first_ex = current_episode[0]
            first_ex_text = []
            partner_persona = []
            your_persona = []
            for line in first_ex['text'].split('\n'):
                # NOTE: we flip "your" and "partner" here since we are taking the 'text'
                # field instead of the 'label'
                if 'partner\'s persona: ' in line:
                    your_persona.append(line.split('partner\'s persona: ')[1])
                elif 'your persona: ' in line:
                    partner_persona.append(line.split('your persona: ')[1])
                else:
                    first_ex_text.append(line)

            your, your_prob, partner, partner_prob = self.get_genders(
                your_persona, partner_persona
            )

            for i, ex in enumerate(current_episode):
                counts['self'][your] += 1
                counts['partner'][partner] += 1
                if i == 0:
                    text = '\n'.join(first_ex_text)
                else:
                    text = ex['text']
                new_ex = {
                    'text': text,
                    'episode_done': True,
                    'your_persona': '\n'.join(your_persona),
                    'partner_persona': '\n'.join(partner_persona),
                    'id': 'ConvAI2 Gender',
                }
                if not self.use_probably:
                    new_ex['partner_prob'] = partner_prob
                    new_ex['your_prob'] = your_prob

                if your is not None and self.labels_to_use != 'partner':
                    # Get the your task
                    labels = [f'SELF:{your}']
                    your_ex = deepcopy(new_ex)
                    your_ex['labels'] = labels
                    your_ex['class_type'] = 'self'
                    new_eps.append(your_ex)

                if partner is not None and self.labels_to_use != 'self':
                    # Get the partner task
                    labels = [f'PARTNER:{partner}']
                    partner_ex = deepcopy(new_ex)
                    partner_ex['labels'] = labels
                    partner_ex['class_type'] = 'partner'
                    new_eps.append(partner_ex)

        if self.labels_to_use == 'all' and self.add_unknown_classes:
            # load about data
            all_about_data = gend_utils.get_inferred_about_data(
                self.opt['task'], self.opt
            )
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                to_samp = int(sample_rate * len(all_about_data))
                sampled = random.sample(all_about_data, to_samp)
                new_eps += sampled
            else:
                new_eps += all_about_data

        if self.is_train:
            random.shuffle(new_eps)

        self.data = new_eps
        print(f'Missing cnt: {self.missing_cnt} / {len(self.data) * 2}')
        for x in ['self', 'partner']:
            print(f'Totals for {x}:')
            subtot = sum(counts[x].values())
            for k, v in counts[x].items():
                print(f'\t{k}: {v} ({v / subtot})')

    def get(self, episode_idx, entry_idx):
        ep = self.data[episode_idx]
        # add label candidates
        # we do not want to hardcode these in the saved data, in case
        # we try to change them
        class_type = ep['class_type']

        if gend_utils.UNKNOWN in ep['labels'][0]:
            # for unknown, we alternate M/F during training
            ep['labels'] = gend_utils.UNKNOWN_LABELS[class_type]

        ep['label_candidates'] = self.label_candidates[class_type]
        return Message(ep)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        shared['persona_map'] = self.persona_map
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import parlai.utils.logging as logging

RESOURCES = [
    DownloadableFile(
        'https://raw.githubusercontent.com/uclanlp/gn_glove/master/wordlist/male_word_file.txt',
        'male_word_file.txt',
        'd431679ce3ef4134647e22cb5fd89e8dbee3f04636f1c7cbae5f28a369acf60f',
        zipped=False,
    ),
    DownloadableFile(
        'https://raw.githubusercontent.com/uclanlp/gn_glove/master/wordlist/female_word_file.txt',
        'female_word_file.txt',
        '5f0803f056de3fbc459589bce26272d3c5453112a3a625fb8ee99c0fbbed5b35',
        zipped=False,
    ),
]


def build(datapath):
    version = 'v1.0'
    dpath = os.path.join(datapath, 'genderation_bias')
    if not build_data.built(dpath, version):
        logging.info('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version)
